import dune.vem
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math, numpy
import scipy
from scipy.sparse.linalg import spsolve
from scipy.sparse import coo_matrix, bmat
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
import dune.fem

from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": 0})
parameters = {"newton.linear.tolerance": 1e-8,
              "newton.tolerance": 5e-6,
              "newton.lineSearch": "simple",
              "newton.linear.verbose": False,
              "newton.verbose": False
              }
Lx,Ly = 1,1.1
orders = [0,1,2]
useVem = True
gridType = "cube" # "voronoi"

def model(spaceU):
    x = SpatialCoordinate(triangle)
    # exact   = (x[0]-Lx)*(x[1]-Ly)*x[0]*x[1]
    exact = sin(2*pi*x[0]/Lx)*sin(3*pi*x[1]/Ly)
    forcing = -div(grad(exact))
    dbc = [dune.ufl.DirichletBC(spaceU, exact, i+1) for i in range(4)]
    return forcing,dbc,exact

def mixed(polyGrid,level,oder, useVem):
    if useVem:
        # unstable spaceU     = dune.fem.space.dglegendre( polyGrid, order=order)
        spaceU     = dune.vem.bbdgSpace( polyGrid, order=order)
        spaceSigma = dune.vem.curlFreeSpace( polyGrid, order=order)
        scheme     = lambda eq: dune.vem.vemScheme(
                       eq, spaceSigma, solver=("suitesparse","umfpack"),
                       parameters=parameters,
                       gradStabilization=None, massStabilization=1)
        diams      = spaceSigma.diameters()
    else:
        # supotimal on cubes
        if gridType.lower() == "cube":
            spaceU     = dune.fem.space.dglegendre( polyGrid, order=order)
        else:
            spaceU     = dune.fem.space.dgonb( polyGrid, order=order)
        spaceSigma = dune.fem.space.raviartThomas( polyGrid, order=order)
        # spaceSigma = dune.fem.space.bdm( polyGrid, order=order+1)
        scheme     = lambda eq: dune.fem.scheme.galerkin(
                       eq, spaceSigma, solver=("suitesparse","umfpack"),
                       parameters=parameters)
        diams      = 2*[2**(-level)]

    forcing,dbc,exact = model(spaceU)
    df  = spaceU.interpolate(exact,name="solution")
    sig = spaceSigma.interpolate(grad(exact),name="sigma")
    # return exact,df,sig,diams

    u     = TrialFunction(spaceU)
    v     = TestFunction(spaceU)
    sigma = TrialFunction(spaceSigma)
    psi   = TestFunction(spaceSigma)
    massOp = inner(sigma,psi) * dx
    gradOp =  u*div(psi) * dx # - inner(dot(sigma,n),exact)
    divOp  = -div(sigma)*v * dx - inner(forcing,v) * dx

    schemeM = scheme(massOp==0)
    schemeG = dune.fem.operator.galerkin( gradOp, spaceU, spaceSigma )
    schemeD = dune.fem.operator.galerkin( divOp, spaceSigma,spaceU )

    M = dune.fem.operator.linear(schemeM).as_numpy
    G = dune.fem.operator.linear(schemeG).as_numpy
    D = dune.fem.operator.linear(schemeD).as_numpy
    # (  M   G  ) ( u     )   ( g ) # g: Dirichlet BC - Neuman BC are natural
    # ( -D   0  ) ( sigma ) = ( f )
    # sigma = -M^{-1}Gu + M^{-1}g
    # f = -Dsigma = DM^{-1}Gu - DM^{-1}g
    # DM^{-1}Gu = f + DM^{-1}g
    # Au = b with A = DM^{-1}G and b = f + DM^{-1}g
    class mixedOp(scipy.sparse.linalg.LinearOperator):
        def __init__(self):
            self.shape = (spaceU.size, spaceU.size)
            self.dtype = df.as_numpy[:].dtype
            self.s1 = spaceSigma.interpolate([0,0],name="tmp").as_numpy[:]
            self.s0 = spaceSigma.interpolate([0,0],name="tmp").as_numpy[:]
            self.y  = spaceU.interpolate(0,name="tmp").as_numpy[:]
        def update(self, x_coeff, f):
            pass
        def _matvec(self, x_coeff):
            x = spaceU.function("tmp", dofVector=x_coeff)
            self.s0[:] = G@x_coeff[:]
            self.s1[:] = spsolve(M,self.s0[:])
            self.y[:] = D@self.s1[:]
            return self.y

    tmp = spaceSigma.interpolate([0,0],name="rhs")
    rhs = spaceU.interpolate(0,name="rhs")
    schemeD(tmp,rhs)
    df.as_numpy[:] = scipy.sparse.linalg.cg(mixedOp(),rhs.as_numpy)[0]
    tmp.as_numpy[:] = -G@df.as_numpy[:]
    sig.as_numpy[:] = spsolve(M,tmp.as_numpy[:])

    vtk = spaceU.gridView.writeVTK("mixed"+"_"+str(level),subsampling=3,
      celldata={"solution":df, "cells":polygons, })
    return exact,df,sig,diams

def primal(polyGrid,level,order, useVem):
    if useVem:
        spaceU = dune.vem.vemSpace( polyGrid, order=order,
                                        testSpaces=[0,order-2,order-2])
        scheme     = lambda eq: dune.vem.vemScheme(
                       eq, spaceU, solver=("suitesparse","umfpack"),
                       parameters=parameters,
                       gradStabilization=1, massStabilization=None)
        diams      = spaceU.diameters()
    else:
        spaceU = dune.fem.space.lagrange( polyGrid, order=order)
        scheme     = lambda eq: dune.fem.scheme.galerkin(
                       eq, spaceU, solver=("suitesparse","umfpack"),
                       parameters=parameters)
        diams      = 2*[2**(-level)]

    forcing,dbc,exact = model(spaceU)
    df = spaceU.interpolate(exact,name="solution")
    # return exact,df,diams

    u      = TrialFunction(spaceU)
    v      = TestFunction(spaceU)
    stiff  = inner(grad(u),grad(v)) * dx - inner(forcing,v) * dx
    scheme = scheme([stiff==0,*dbc])

    scheme.solve(target=df)
    vtk = spaceU.gridView.writeVTK("primal"+"_"+str(level),subsampling=3,
      celldata={"solution":df, "cells":polygons, })
    return exact,df,diams

for order in orders:
    oldErrors = []
    oldDiams = None
    for i in range(1,6):
        errors = []
        sizes = []
        N = 2**(i+1)
        if gridType.lower() == "cube":
            polyGrid = dune.vem.polyGrid(
                  cartesianDomain([0.,0.],[Lx,Ly],[N,N]), cubes=True
              )
        elif gridType.lower() == "simplex":
            assert useVem, "RT not working for simplex grids"
            polyGrid = dune.vem.polyGrid(
                  cartesianDomain([0.,0.],[Lx,Ly],[N,N]), cubes=False
              )
        elif gridType.lower() == "voronoi":
            assert useVem
            polyGrid = dune.vem.polyGrid(
                  dune.vem.voronoiCells([[0,0],[Lx,Ly]], N*N, lloyd=200, fileName="test", load=True)
              )
        indexSet = polyGrid.indexSet
        @gridFunction(polyGrid, name="cells")
        def polygons(en,x):
            return polyGrid.hierarchicalGrid.agglomerate(indexSet.index(en))

        #######################################################
        exact,df,sigma,diam = mixed(polyGrid,i,order, useVem)
        e  = exact-df
        de = grad(exact)-sigma
        err = [e**2,inner(de,de),div(de)**2]
        errors += [ numpy.sqrt(e) for e in integrate(polyGrid, err, order=8) ]
        sizes += [ df.size, sigma.size ]
        #######################################################
        exact,df,diam = primal(polyGrid,i,order+1, useVem)
        e = exact-df
        de = grad(e)
        d2e = grad(grad(e))
        err = [e**2,inner(de,de),div(de)**2,inner(d2e,d2e)]
        errors += [ numpy.sqrt(e) for e in integrate(polyGrid, err, order=8) ]
        sizes += [ df.size ]
        #######################################################

        print(order,i,*sizes,*errors,"# order=",order,diam,flush=True)
        if len(oldErrors)>0:
            factor = oldDiams[0] / diam[0]
            print(order,i,*[ math.log(oe/e)/math.log(factor)
                    for oe,e in zip(oldErrors,errors) ],
                    "#eoc order=",order, flush=True)
        oldErrors = errors
        oldDiams = diam



from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.fem.operator import linear as linearOperator

import ufl.algorithms
from ufl import *
import dune.ufl

from script import runTest, checkEOC, interpolate
from interpolate import interpolate_fourthorder

dimR = 1

parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "jacobi",
              "penalty": 40,  # for the bbdg scheme
              "newton.linear.verbose": False,
              "newton.verbose": True
              }

def varyingcoeff(space, exact):
    laplace = lambda w: div(grad(w))
    H = lambda w: grad(grad(w))
    u = TrialFunction(space)
    v = TestFunction(space)
    x = SpatialCoordinate(space)

    laplaceCoeff = 1
    mu           = 1

    kappa = 1./(1+dot(x,x))
    beta  = exp(-x[0]*x[1])
    gamma = sin(dot(x,x))**2

    a = ( kappa*inner(H(u[0]),H(v[0])) +\
          laplaceCoeff*beta*inner(grad(u),grad(v)) +\
          mu*gamma*inner(u,v)
        ) * dx

    # right hand side and the boundary conditions
    q = sum([ H(kappa*H(exact[0])[i,j])[i,j] for i in range(2) for j in range(2) ])
    b = ( q -\
          laplaceCoeff*div(beta*grad(exact[0])) +\
          mu*gamma*exact[0] ) * v[0] * dx
    dbc = [dune.ufl.DirichletBC(space, [0], i+1) for i in range(4)]

    biLaplaceCoeff = kappa
    diffCoeff      = laplaceCoeff*beta
    massCoeff      = mu*gamma

    scheme = dune.vem.vemScheme([a==b, *dbc], space,
                            solver="cg",
                            hessStabilization=biLaplaceCoeff,
                            gradStabilization=diffCoeff,
                            massStabilization=massCoeff,
                            parameters=parameters)

    df = discreteFunction(space, name="solution")
    info = scheme.solve(target=df)

    edf = exact-df
    err = [inner(edf,edf),
            inner(grad(edf),grad(edf)),
            inner(grad(grad(edf)),grad(grad(edf)))]

    return err

def runTestVaryingcoeff(testSpaces, order):
      x = SpatialCoordinate(triangle)
      exact = as_vector( dimR*[sin(pi*x[0])**2*sin(pi*x[1])**2] )
      spaceConstructor = lambda grid, r: dune.vem.vemSpace( grid,
                                                            order=order,
                                                            dimRange=r,
                                                            testSpaces=testSpaces )

      if (order==2):
        if len(testSpaces[0]) == 2: # conforming space seems to have worse L^2 eoc?
            expected_eoc = [order+1, order, order-1]
        else:
            expected_eoc = [order, order, order-1]
      else:
        expected_eoc = [order+1, order, order-1]

      if (interpolate()):
            eoc = runTest(exact, spaceConstructor, interpolate_fourthorder)
      else:
            eoc = runTest(exact, spaceConstructor, varyingcoeff)

      return eoc, expected_eoc

def main():
      ret = 0
      orders = [2,3]
      for order in orders:
            print("order: ", order)
            C1NCtestSpaces = [ [0], [order-3,order-2], [order-4] ]
            print("C1 non conforming test spaces: ", C1NCtestSpaces)

            eoc, expected_eoc = runTestVaryingcoeff( C1NCtestSpaces, order )

            ret += checkEOC(eoc, expected_eoc)

            print("order: ", order)
            C1ConftestSpaces = [ [0,0], [order-4,order-3], [order-4] ]
            print("C1 conforming test spaces: ", C1ConftestSpaces)

            eoc, expected_eoc = runTestVaryingcoeff( C1ConftestSpaces, order )

            ret += checkEOC(eoc, expected_eoc)
      assert ret>0, "some test went wrong"

main()



from ufl import *
import dune.ufl

parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "jacobi",
              "penalty": 40,  # for the dg schemes
              "newton.linear.verbose": False,
              "newton.verbose": False
              }

def elliptic(space, exact):
    # set up scheme, df, and solve
    u = TrialFunction(space)
    v = TestFunction(space)
    x = SpatialCoordinate(space)

    massCoeff = 1+sin(dot(x,x))       # factor for mass term
    diffCoeff = 1-0.9*cos(dot(x,x))

    a = (diffCoeff*inner(grad(u),grad(v)) + massCoeff*dot(u,v) ) * dx

    # finally the right hand side and the boundary conditions
    b = (-div(diffCoeff*grad(exact[0])) + massCoeff*exact[0] ) * v[0] * dx
    dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]

    df = space.interpolate([0],name="solution")
    scheme = dune.vem.vemScheme( [a==b, *dbc], space, solver="cg",
                            gradStabilization=diffCoeff,
                            massStabilization=massCoeff,
                            parameters=parameters )

    info = scheme.solve(target=df)

    edf = exact-df
    err = [inner(edf,edf), inner(grad(edf),grad(edf))]

    return err


# %% [markdown]
# .. index:: Solvers; Saddle Point (Uzawa)
#
# # Saddle Point Solver (using Scipy)
# %%
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import numpy
from scipy.sparse import linalg
from dune.grid import cartesianDomain
from dune.alugrid import aluCubeGrid
from ufl import SpatialCoordinate, CellVolume, TrialFunction, TestFunction,\
                inner, dot, div, nabla_grad, grad, dx, as_vector, transpose, Identity, sym, \
                FacetNormal, ds, dS, avg, jump, CellVolume, FacetArea, conditional
from dune.ufl import Constant, DirichletBC
import dune.fem

import dune.vem
from dune.fem.operator import galerkin as galerkinOperator
from dune.vem import vemScheme as vemScheme
from dune.fem.scheme import galerkin as galerkinScheme

def plot(target):
    fig = pyplot.figure(figsize=(10,10))
    target[0].plot(colorbar="vertical", figure=(fig, 211))
    target[1].plot(colorbar="vertical", figure=(fig, 212))
    pyplot.show()

def dgLaplace(beta, p,q, spc, p_bnd, precondDbnd):
    n             = FacetNormal(spc)
    he            = avg( CellVolume(spc) ) / FacetArea(spc)
    hbnd          = CellVolume(spc) / FacetArea(spc)
    aInternal     = dot(grad(p), grad(q)) * dx
    diffSkeleton  = beta/he*jump(p)*jump(q)*dS -\
                    dot(avg(grad(p)),n('+'))*jump(q)*dS -\
                    jump(p)*dot(avg(grad(q)),n('+'))*dS
    for cond in precondDbnd:
        diffSkeleton  += ( beta/hbnd * (p-p_bnd) -
                           dot(grad(p),n) ) * q * conditional(cond,1,0) * ds
    return aInternal + diffSkeleton
class Uzawa:
    def __init__(self, grid,
                 spcU, spcP, dbc_u,
                 mainModel, gradModel, divModel,
                 mu, nu,
                 tolerance=1e-9, verbose=False,
                 lagrange=False, precondDbnd=None):
        self.dimension = grid.dimension
        self.verbose = verbose
        self.tolerance2 = tolerance**2
        self.dbc_u = dbc_u
        self.mu = mu
        self.nu = nu
        p = TrialFunction(spcP)
        q = TestFunction(spcP)
        massModel   = p*q * dx
        preconModel = inner(grad(p),grad(q)) * dx

        if not lagrange:
            self.mainOp = vemScheme( ( mainModel==0, *dbc_u ),
                                       gradStabilization=as_vector([self.mu*2,self.mu*2]),
                                       massStabilization=as_vector([self.nu,self.nu])
                                     )
        else:
            self.mainOp = galerkinScheme( ( mainModel==0, *dbc_u ) )
        gradOp    = galerkinOperator( gradModel, spcP,spcU)
        divOp     = galerkinOperator( divModel, spcU,spcP)
        massOp    = galerkinOperator( massModel, spcP)

        self.mainLinOp = self.mainOp.linear()
        self.G    = gradOp.linear().as_numpy
        self.D    = divOp.linear().as_numpy
        self.M    = massOp.linear().as_numpy
        self.Minv = lambda rhs: linalg.spsolve(self.M,rhs)

        if self.mainOp.model.nu > 0:
            print("adding outer preconditioner",
                  "using nu=",self.mainOp.model.nu,
                  "and mu=",self.mainOp.model.mu,flush=True)
            if not lagrange:
                o = spcP.order
                preconModel = dgLaplace(10*(o+1)**2, p,q, spcP, 0, precondDbnd)
                preconOp    = galerkinOperator(preconModel, spcP)
            else:
                x = SpatialCoordinate(spcP)
                preconModel = inner(grad(p),grad(q)) * dx
                preconOp    = galerkinOperator([ preconModel ]
                              + [ DirichletBC(spcP,0,cond) for cond in precondDbnd ])
            self.P    = preconOp.linear().as_numpy.tocsc()
            self.Pinv = linalg.splu(self.P)
        else:
            self.Pinv = None
            print("No outer preconditioner",
                  "with nu=",self.mainOp.model.nu,
                  "and mu=",self.mainOp.model.mu,flush=True)

        self.rhsVelo  = spcU.interpolate(spcU.dimRange*[0], name="vel")
        self.rhsPress = spcP.interpolate(0, name="pres")
        self.rhs_u  = self.rhsVelo.as_numpy
        self.rhs_p  = self.rhsPress.as_numpy
        self.r      = numpy.copy(self.rhs_p)
        self.d      = numpy.copy(self.rhs_p)
        self.precon = numpy.copy(self.rhs_p)
        self.xi     = numpy.copy(self.rhs_u)

    def solve(self,target):
        velocity = target[0]
        pressure = target[1]
        info = {"uzawa.outer.iterations":0,
                "uzawa.converged":False}
        # problem is linear but coefficients depend on previous time step so need to reassemble
        self.mainOp.jacobian(velocity, self.mainLinOp)
        A = self.mainLinOp.as_numpy.tocsc()
        Ainv = linalg.splu(A)
        sol_u = velocity.as_numpy
        sol_p = pressure.as_numpy
        # right hand side for Shur complement problem
        velocity.clear()
        self.mainOp(velocity,self.rhsVelo)
        self.rhs_u *= -1
        self.xi[:]  = self.G*sol_p
        self.rhs_u -= self.xi
        self.mainOp.setConstraints(self.rhsVelo)
        sol_u[:]      = Ainv.solve(self.rhs_u[:])
        self.rhs_p[:] = self.D*sol_u
        self.r[:]     = self.Minv(self.rhs_p[:])
        if self.Pinv:
            self.precon.fill(0)
            self.precon[:] = self.Pinv.solve(self.rhs_p[:])
            self.r *= self.mu.value
            self.r += self.nu.value*self.precon
        self.d[:] = self.r[:]
        delta = numpy.dot(self.r,self.rhs_p)
        if self.verbose:
            print(0,delta,self.tolerance2)
        # cg type iteration
        for m in range(1,1000):
            self.xi.fill(0)
            self.rhs_u[:] = self.G*self.d
            self.mainOp.setConstraints([0,]*self.dimension, self.rhsVelo)
            self.xi[:] = Ainv.solve(self.rhs_u[:])
            self.rhs_p[:] = self.D*self.xi
            rho = delta / numpy.dot(self.d,self.rhs_p)
            sol_p += rho*self.d
            sol_u -= rho*self.xi
            self.rhs_p[:] = self.D*sol_u
            self.r[:] = self.Minv(self.rhs_p[:])
            if self.Pinv:
                self.precon.fill(0)
                self.precon[:] = self.Pinv.solve(self.rhs_p[:])
                self.r *= self.mu.value
                self.r += self.nu.value*self.precon
            oldDelta = delta
            delta = numpy.dot(self.r,self.rhs_p)
            if self.verbose:
                print(m,delta,self.tolerance2)
            if delta < self.tolerance2:
                info["uzawa.converged"] = True
                break
            gamma = delta/oldDelta
            self.d *= gamma
            self.d += self.r
        info["uzawa.outer.iterations"] = m
        return info



# %% [markdown]
# # 2D navier stokes around a cylinder
#
# https://hplgit.github.io/fenics-tutorial/pub/sphinx1/._ftut1004.html#flow-past-a-cylinder

# %%
import matplotlib

matplotlib.rc("image", cmap="jet")
from math import sqrt
import matplotlib.pyplot as plt
import pygmsh
from dune.ufl import Constant, DirichletBC
import dune.fem.space
import dune.fem.scheme
import dune.vem
from uzawa import Uzawa, plot
from ufl import dot, nabla_grad, SpatialCoordinate, as_vector

explicit = False
useVem = True
order = 2
T   = 5.0  # final time
# tau = T / 10000
# tau = T / 5000
tau = T / 4000


with pygmsh.occ.Geometry() as geom:
    # geom.characteristic_length_max = 0.02
    geom.set_mesh_size_callback( lambda dim, tag, x, y, z, lc:
          min(0.04+(x-0.2)**2+(y-0.2)**2 - 0.1**2, 0.1) ) # p=4
          # min(0.01+(x-0.2)**2+(y-0.2)**2 - 0.1**2, 0.1) ) # p=2
          #  min(0.005+(x-0.2)**2+(y-0.2)**2 - 0.05**2, 0.02) )
    rectangle = geom.add_rectangle([0, 0, 0], 2.2, 0.41)
    cylinder = geom.add_disk([0.2, 0.2, 0.0], 0.05)
    geom.boolean_difference(rectangle, cylinder)

    mesh = geom.generate_mesh()
    points, cells = mesh.points, mesh.cells_dict
    domain = {
        "vertices": points[:, :2].astype(float),
        "simplices": cells["triangle"].astype(int),
    }
    print(len(domain["simplices"]))

gridView = dune.vem.polyGrid(domain)
# gridView.plot()
# plt.show()
if useVem:
    if explicit:
        path="uzawaeVem_p" + str(order)
    else:
        path="uzawaVem_p" + str(order)
    space_u = dune.vem.divFreeSpace( gridView, order=order)
    space_p = dune.fem.space.finiteVolume( gridView )
else:
    if explicit:
        path="uzawaeTH_p" + str(order)
    else:
        path="uzawaTH_p" + str(order)
    space_u = dune.fem.space.lagrange( gridView, order=order, dimRange=2)
    space_p = dune.fem.space.lagrange( gridView, order=order-1 )

mu  = 0.001

u_h = space_u.interpolate([0, 0], name="u_h")
u_h_n = space_u.interpolate([0, 0], name="u_h")
p_h = space_p.interpolate(0, name="p_h")

x = SpatialCoordinate(space_u)
inflow = [6 * x[1] * (0.41 - x[1]) / 0.41**2, 0] # 6->18
dbc_u_in = DirichletBC(space_u, inflow, x[0] <= 1e-8)
dbc_u_noflow = DirichletBC(space_u, [0, 0], None)
dbc_u_out = DirichletBC(space_u, [None,None], 2.2 - x[0] <= 1e-8)
dbc_p_out = DirichletBC(space_p, 0, 2.2 - x[0] <= 1e-8)

# %%
from pathlib import Path
p = Path(path)
p.mkdir(exist_ok=True)
vtk = gridView.sequencedVTK(path+f"/2DCylinder",
              subsampling=3,
              pointdata=[p_h], pointvector=[u_h])
vtk()

time = 0
saveStep = 0.05
nextSave = saveStep

uzawa = Uzawa(gridView, space_u, space_p,
              [dbc_u_in,dbc_u_out,dbc_u_noflow], as_vector([0,0]),
              mu, tau, u_h_n,
              tolerance=1e-6, precondition=True, verbose=False,
              lagrange=not useVem, explicit=explicit)
while time < T:
    u_h_n.assign(u_h)

    info = uzawa.solve([u_h,p_h])

    time += tau
    print(time, info, flush=True)
    if time > nextSave:
        vtk()
        nextSave += saveStep



import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import numpy as np
import pandas as pd
import math, argparse
from scipy.sparse.linalg import spsolve
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem.operator import linear as linearOperator
import dune.vem

from ufl import *
import dune.ufl

parser = argparse.ArgumentParser(description='set order and stabilization')
parser.add_argument('-L', type=int, required=True, help="max-level > 0")
parser.add_argument('-l', type=int, required=True, help="polynomial order >= 1")
parser.add_argument('-s', type=int, required=True, help="use stabilization = [-1|0|1]")
parser.add_argument('-p', type=str, required=True, help="problem = [laplace|biharmonic]")

# Space and stabilization is influenced by the '-s' parameter:
#    1: normal space with stabilization
#    0: normal space w.o. stabilization
#   -1: extended space w.o. stabilization
args = parser.parse_args()
useStab       = args.s     # one of 1,0,-1
maxLevel      = args.L     # number of levels to use for EOC computation
problem       = args.p     # problem to solve (laplace or biharmonic)
order         = args.l     # order (>=1 for laplace >=2 for biharmonic)

if useStab == -1: # set range space of value, gradient, and hessian projections to P_l,P_l, and P_{l-1}
    orders = [order,order,order-1]
else:             # default setup
    orders = [order,order-1,order-2]

if problem == "biharmonic": # parameters for H^2-conforming space
    spaceArgs = {"order":orders, "testSpaces":[ [0,0], [order-4,order-3], [order-4] ] }
else: # parameters for H^1-conforming space
    spaceArgs = {"order":orders, "testSpaces":[ [0,-1], [order-2,-1], [order-2] ] }

# Now we define the model starting with the exact solution and boundary conditions
uflSpace = dune.ufl.Space(2)
x = SpatialCoordinate(uflSpace)
exact = sin(2*pi*x[0])**2*sin(2*pi*x[1])**2
dbc = [dune.ufl.DirichletBC(uflSpace, 0, i+1) for i in range(4)]

# and now the bilinear and linear forms
u = TrialFunction(uflSpace)
v = TestFunction(uflSpace)
H = lambda w: grad(grad(w))

kappa = 10/(1e-2+dot(x,x))

if problem == "biharmonic":
    a = kappa*inner(H(u),H(v)) * dx
    b = sum([ H(kappa*H(exact)[i,j])[i,j] for i in range(2) for j in range(2) ]) * v * dx
else:
    a = kappa*inner(grad(u),grad(v)) * dx
    b = - div(kappa*grad(exact)) * v * dx

# Finally the stabilization depending on provided runtime arguments
if useStab == 1:
    print("# original space with standard stabilization!")
    stabCoeff = kappa
elif useStab == 0:
    print("# no stabilization with original space!")
    stabCoeff = None
else: # useStab=-1
    print("# using extended projections with no stabilization!")
    stabCoeff = None

# We now define a function to compute the solution and the $L^2,H^1$ error
# given a grid and a space
def compute(grid, space):
    df = discreteFunction(space, name="solution") # space.interpolate([0],name="solution")
    scheme = dune.vem.vemScheme([a==b, *dbc], space,
                        hessStabilization=stabCoeff if problem=="biharmonic" else None,
                        gradStabilization=stabCoeff if problem=="laplace" else None,
                        massStabilization=None)
    A = dune.fem.operator.linear(scheme).as_numpy
    rhs = df.copy()
    scheme(space.zero,rhs)
    rhs *= -1
    df.as_numpy[:] = spsolve(A, rhs.as_numpy[:])
    info = {"linear_iterations":"direct"}
    edf = exact-df
    energy  = replace(a,{u:edf,v:edf}).integrals()[0].integrand()
    err = [energy,
           inner(edf,edf),
           inner(grad(edf),grad(edf))]
    if problem == "biharmonic":
        err += [inner(grad(grad(edf)),grad(grad(edf)))]
    errors  = [ math.sqrt(e) for e in integrate(grid, err, order=8) ]
    return df, errors, info

# Finally we iterate over the requested methods and solve the problems
results = []
for level in range(maxLevel+1):
    constructor = cartesianDomain([0,0],[1,1],[4*2**level,4*2**level])
    polyGrid = dune.vem.polyGrid(constructor, cubes=True)
    res = []
    space = dune.vem.vemSpace(polyGrid, **spaceArgs)
    dfs,errors,info = compute(polyGrid, space)
    results += [ [polyGrid.hierarchicalGrid.agglomerate.size,space.size,info["linear_iterations"],*errors] ]

# compute EOCs and output results
h = lambda size: 1/sqrt(size)
eoc = [(len(results[0])-3)*[0]]
for level in range(len(results)):
    if level<len(results)-1:
        eoc += [ [math.log(results[level+1][3+j]/results[level][3+j])/
                 math.log(h(results[level+1][0])/h(results[level][0]))
                 for j in range(len(eoc[:][0]))] ]

results = np.array(results)
eoc = np.array(eoc)
keys = {'grid-size': results[:,0],
        "energy-error": results[:,3],
        "E-eoc": eoc[:,0],
        'L2-error': results[:,4],
        "L2-eoc": eoc[:,1],
        'H1-error': results[:,5],
        "H1-eoc": eoc[:,2]
       }
if problem == "biharmonic":
    keys['H2-error'] = results[:,6]
    keys["H2-eoc"] = eoc[:,3]
columns = list(keys.keys())
table = pd.DataFrame(keys, index=range(len(results)), columns=columns)
for c in columns:
    if "eoc" in c:
        table[c] = table[c].apply('{:.2f}'.format)
print(table)



from __future__ import division, print_function, unicode_literals

from dune.ufl.codegen import generateMethod
from ufl import SpatialCoordinate, Coefficient, replace, diff, as_vector, MaxCellEdgeLength
from ufl.constantvalue import Zero
from ufl.core.expr import Expr
from ufl.tensors import ListTensor
from dune.source.cplusplus import Variable, UnformattedExpression, AccessModifier, maxEdgeLength
from ufl.algorithms import expand_compounds, expand_derivatives, expand_indices, expand_derivatives

def codeVEM(self, name, targs):
    code = self._code(name,targs)

    u = self.trialFunction
    ubar = Coefficient(u.ufl_function_space())
    mStab = self.mStab
    if isinstance(mStab,Expr):
        if mStab.ufl_shape == (1,):
            mStab = as_vector(u.ufl_shape[0]*[mStab[0]])
        try:
            mStab = expand_indices(expand_derivatives(expand_compounds(mStab)))
        except:
            pass
        dmStab = replace(
                   expand_derivatives( diff(replace(mStab,{u:ubar}),ubar) ),
                   {ubar:u} )
        assert mStab.ufl_shape == u.ufl_shape
        dmStab = as_vector([
                    replace(
                      expand_derivatives( diff(replace(mStab,{u:ubar}),ubar) )[i,i],
                    {ubar:u} )
                 for i in range(u.ufl_shape[0]) ])
    else:
        dmStab = None

    gStab = self.gStab
    if isinstance(gStab,Expr):
        if gStab.ufl_shape == (1,):
            gStab = as_vector(u.ufl_shape[0]*[gStab[0]])
        try:
            gStab = expand_indices(expand_derivatives(expand_compounds(gStab)))
        except:
            pass
        assert gStab.ufl_shape == u.ufl_shape
        dgStab = as_vector([
                    replace(
                      expand_derivatives( diff(replace(gStab,{u:ubar}),ubar) )[i,i],
                    {ubar:u} )
                 for i in range(u.ufl_shape[0]) ])
    else:
        dgStab = None

    hStab = self.hStab
    if isinstance(hStab,Expr):
        if hStab.ufl_shape == (1,):
            hStab = as_vector(u.ufl_shape[0]*[hStab[0]])
        try:
            hStab = expand_indices(expand_derivatives(expand_compounds(hStab)))
        except:
            pass
        assert hStab.ufl_shape == u.ufl_shape
        dhStab = as_vector([
                    replace(
                      expand_derivatives( diff(replace(hStab,{u:ubar}),ubar) )[i,i],
                    {ubar:u} )
                 for i in range(u.ufl_shape[0]) ])
    else:
        dhStab = None

    code.append(AccessModifier("public"))
    x = SpatialCoordinate(self.space.cell())
    maxCellEdgeLength = MaxCellEdgeLength(self.space.cell())
    predefined = {}
    # self.predefineCoefficients(predefined, False)
    spatial = Variable('const auto', 'y')
    predefined[x] = UnformattedExpression('auto', 'entity().geometry().global( Dune::Fem::coordinate( x ) )')
    predefined[maxCellEdgeLength] = maxEdgeLength(self.cellGeometry())
    self.predefineCoefficients(predefined)
    generateMethod(code, hStab,
            'RRangeType', 'hessStabilization',
            args=['const Point &x',
                  'const DRangeType &u'],
            targs=['class Point','class DRangeType'], static=False, const=True,
            predefined=predefined)
    generateMethod(code, dhStab,
            'RRangeType', 'linHessStabilization',
            args=['const Point &x',
                  'const DRangeType &u'],
            targs=['class Point','class DRangeType'], static=False, const=True,
            predefined=predefined)
    generateMethod(code, gStab,
            'RRangeType', 'gradStabilization',
            args=['const Point &x',
                  'const DRangeType &u'],
            targs=['class Point','class DRangeType'], static=False, const=True,
            predefined=predefined)
    generateMethod(code, dgStab,
            'RRangeType', 'linGradStabilization',
            args=['const Point &x',
                  'const DRangeType &u'],
            targs=['class Point','class DRangeType'], static=False, const=True,
            predefined=predefined)
    generateMethod(code, mStab,
            'RRangeType', 'massStabilization',
            args=['const Point &x',
                  'const DRangeType &u'],
            targs=['class Point','class DRangeType'], static=False, const=True,
            predefined=predefined)
    generateMethod(code, dmStab,
            'RRangeType', 'linMassStabilization',
            args=['const Point &x',
                  'const DRangeType &u'],
            targs=['class Point','class DRangeType'], static=False, const=True,
            predefined=predefined)
    return code

def transform(space,hStab,gStab,mStab):
    if not type(hStab) in [list,tuple,ListTensor,Zero]: # note __getitem__ exists for any ufl expressions
        hStab = [hStab]
    if not type(gStab) in [list,tuple,ListTensor,Zero]: # note __getitem__ exists for any ufl expressions
        gStab = [gStab]
    if not type(mStab) in [list,tuple,ListTensor,Zero]: # note __getitem__ exists for any ufl expressions
        mStab = [mStab]
    exprs = []
    baseSignature = []
    if mStab[0] is not None:
        baseSignature += [["mass",mStab]]
        exprs += [x for x in mStab if not type(x) in [int,float]]
    if gStab[0] is not None:
        baseSignature += [["grad",gStab]]
        exprs += [x for x in gStab if not type(x) in [int,float]]
    if hStab[0] is not None:
        baseSignature += [["hess",hStab]]
        exprs += [x for x in hStab if not type(x) in [int,float]]
    def transform_(model):
        model.baseSignature = baseSignature
        if model.baseName == "vemintegrands":
            return
        model._code = model.code
        model.code  = lambda *args,**kwargs: codeVEM(model,*args,**kwargs)
        model.space = space
        model.hStab = as_vector(hStab) if hStab[0] else hStab
        model.gStab = as_vector(gStab) if gStab[0] else gStab
        model.mStab = as_vector(mStab) if mStab[0] else mStab
        model.baseName = "vemintegrands"
    return [transform_, exprs]



#!/usr/bin/python

# Find the minimum-area bounding box of a set of 2D points
#
# The input is a 2D convex hull, in an Nx2 numpy array of x-y co-ordinates.
# The first and last points points must be the same, making a closed polygon.
# This program finds the rotation angles of each edge of the convex polygon,
# then tests the area of a bounding box aligned with the unique angles in
# 90 degrees of the 1st Quadrant.
# Returns the
#
# Tested with Python 2.6.5 on Ubuntu 10.04.4
# Results verified using Matlab

# Copyright (c) 2013, David Butterworth, University of Queensland
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys  # maxint
import numpy as np

def minBoundingRect(hull_points_2d):
    # Compute edges (x2-x1,y2-y1)
    edges = np.zeros( (len(hull_points_2d)-1,2) ) # empty 2 column array
    for i in range( len(edges) ):
        edge_x = hull_points_2d[i+1,0] - hull_points_2d[i,0]
        edge_y = hull_points_2d[i+1,1] - hull_points_2d[i,1]
        edges[i] = [edge_x,edge_y]

    # Calculate edge angles   atan2(y/x)
    edge_angles = np.zeros( (len(edges)) ) # empty 1 column array
    for i in range( len(edge_angles) ):
        edge_angles[i] = np.arctan2( edges[i,1], edges[i,0] )

    # Check for angles in 1st quadrant
    for i in range( len(edge_angles) ):
        edge_angles[i] = abs( edge_angles[i] % (np.pi/2) ) # want strictly positive answers

    # Remove duplicate angles
    edge_angles = np.unique(edge_angles)

    # Test each angle to find bounding box with smallest area
    min_bbox = (0, sys.maxsize, 0, 0, 0, 0, 0, 0) # rot_angle, area, width, height, min_x, max_x, min_y, max_y
    for i in range( len(edge_angles) ):

        # Create rotation matrix to shift points to baseline
        # R = [ cos(theta)      , cos(theta-PI/2)
        #       cos(theta+PI/2) , cos(theta)     ]
        R = np.array([ [ np.cos(edge_angles[i]), np.cos(edge_angles[i]-(np.pi/2)) ],
                       [ np.cos(edge_angles[i]+(np.pi/2)), np.cos(edge_angles[i]) ] ])

        # Apply this rotation to convex hull points
        rot_points = np.dot(R, np.transpose(hull_points_2d) ) # 2x2 * 2xn

        # Find min/max x,y points
        min_x = np.nanmin(rot_points[0], axis=0)
        max_x = np.nanmax(rot_points[0], axis=0)
        min_y = np.nanmin(rot_points[1], axis=0)
        max_y = np.nanmax(rot_points[1], axis=0)

        # Calculate height/width/area of this bounding rectangle
        width = max_x - min_x
        height = max_y - min_y
        area = width*height

        # Store the smallest rect found first (a simple convex hull might have 2 answers with same area)
        if (area < min_bbox[1]):
            min_bbox = ( edge_angles[i], area, width, height, min_x, max_x, min_y, max_y )
        # Bypass, return the last found rect
        #min_bbox = ( edge_angles[i], area, width, height, min_x, max_x, min_y, max_y )

    # Re-create rotation matrix for smallest rect
    angle = min_bbox[0]
    R = np.array([ [ np.cos(angle), np.cos(angle-(np.pi/2)) ],
                   [ np.cos(angle+(np.pi/2)), np.cos(angle) ] ])

    # Project convex hull points onto rotated frame
    proj_points = np.dot(R, np.transpose(hull_points_2d) ) # 2x2 * 2xn

    # min/max x,y points are against baseline
    min_x = min_bbox[4]
    max_x = min_bbox[5]
    min_y = min_bbox[6]
    max_y = min_bbox[7]

    # Calculate center point and project onto rotated frame
    center_x = (min_x + max_x)/2
    center_y = (min_y + max_y)/2
    center_point = np.dot( [ center_x, center_y ], R )

    # Calculate corner points and project onto rotated frame
    corner_points = np.zeros( (4,2) ) # empty 2 column array
    corner_points[0] = np.dot( [ max_x, min_y ], R )
    corner_points[1] = np.dot( [ min_x, min_y ], R )
    corner_points[2] = np.dot( [ min_x, max_y ], R )
    corner_points[3] = np.dot( [ max_x, max_y ], R )

    return (angle, min_bbox[1], min_bbox[2], min_bbox[3], center_point, corner_points) # rot_angle, area, width, height, center_point, corner_points



import dune.vem
import math, random
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.vem import voronoiCells, vemScheme, vemSpace, polyGrid
import numpy

import ufl.algorithms
from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": -1})

parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "jacobi",
              "penalty": 40,  # for the dg schemes
              "newton.linear.verbose": False,
              "newton.verbose": False
              }

def test(exact, spaceConstructor, get_df):
    Nval = 5
    N_values = [2**i*Nval for i in range(4)]
    results = []
    for N in N_values:
        print('grid size N:', N)
        # set up grid for testing
        constructor = cartesianDomain([0,0],[1,1],[N,N])
        cells = voronoiCells(constructor,N*N,"voronoiseeds",load=True,show=False,lloyd=10)
        grid = create.grid("agglomerate", cells, convex=True)

        # grid = dune.vem.polyGrid( dune.vem.voronoiCells([[-0.5,-0.5],[1,1]], 50, lloyd=100) )

        # get dimension of range
        dimRange = exact.ufl_shape[0]
        print('dim Range:', dimRange)

        # construct space to use
        space = spaceConstructor(grid,dimRange)

        # get sizes for eoc calculation
        sizes = [grid.hierarchicalGrid.agglomerate.size, space.size, *space.diameters()]

        # solve
        df = get_df(space)
        df.plot(level=3)
        plot(df-exact, grid=grid, gridLines=None, level=3)

        # calculate errors
        edf = exact-df
        err = [inner(edf,edf), inner(grad(edf),grad(edf))]
        errors = [ numpy.sqrt(e) for e in integrate(grid, err, order=8) ]
        # print(errors)

        results += [[sizes] + [errors]]

    # calculate eocs
    h = lambda sizes: sizes[3]
    eoc = len(results[0][1])*[-1]
    for level in range(0,len(N_values)):
        print(*results[level], *eoc)
        if level < len(N_values) -1:
            eoc = [math.log(results[level+1][1][j]/results[level][1][j])/
            math.log(h(results[level+1][0])/h(results[level][0]))
            for j in range(len(eoc))]

def runTest(exact, spaceConstructor):
    # interpolation test
    # test( exact, spaceConstructor, lambda space: space.interpolate([0],name="solution") )

    # elliptic test
    def testElliptic(space):
        # set up scheme, df, and solve
        u = TrialFunction(space)
        v = TestFunction(space)
        x = SpatialCoordinate(space)

        massCoeff = 1+sin(dot(x,x))       # factor for mass term
        diffCoeff = 11-0.9*cos(dot(x,x))

        a = (diffCoeff*inner(grad(u),grad(v)) + massCoeff*dot(u,v) ) * dx

        # finally the right hand side and the boundary conditions
        b = (-div(diffCoeff*grad(exact[0])) + massCoeff*exact[0] ) * v[0] * dx
        dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]

        df = space.interpolate([0],name="solution")
        scheme = dune.vem.vemScheme( [a==b, *dbc], space, solver="cg",
                             gradStabilization=diffCoeff,
                             massStabilization=massCoeff,
                             parameters=parameters )


        info = scheme.solve(target=df)
        return df

    test( exact, spaceConstructor, testElliptic )

def main():
    # test with conforming second order VEM space
    order=3
    conformingSpaceConstructor = lambda grid, r: dune.vem.vemSpace(grid, order=order, dimRange=r, testSpaces = [-1,order-1,order-2])
    # define exact solution
    x = ufl.SpatialCoordinate(ufl.triangle)
    exact = as_vector( [x[0]*x[1] * cos(pi*x[0]*x[1])] )

    runTest(exact, conformingSpaceConstructor)

main()


# <markdowncell>
# # Laplace problem
#
# We first consider a simple Laplace problem with Dirichlet boundary conditions
# \begin{align*}
#   -\Delta u &= f, && \text{in } \Omega, \\
#           u &= g, && \text{on } \partial\Omega,
# \end{align*}
# with $\Omega=[-\frac{1}{2},1]^2$ and choosing the forcing and the boundary conditions
# so that the exact solution is equal to
# \begin{align*}
#   u(x,y) &= xy\cos(\pi xy)
# \end{align*}

# First some setup code:
# <codecell>
try:
    import dune.vem
except:
    import sys
    sys.exit(0)
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math, pickle, gc
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate
from dune.fem import parameter
from dune.vem import voronoiCells
from dune.vem import space as generalSpace

from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": 0})

# <markdowncell>
# we can compare different method, e.g., a lagrange space (on the # subtriangulation),
# a bounding box dg space and a conforming/non conforming VEM space

# <codecell>
massCoeff = 0   # 1+sin(dot(x,x))
diffCoeff = 1   # -0.9*cos(dot(x,x))
basisChoice=2
orders = range(1,12)
resolutions = [4,8,16,32] # [2*2**n for n in range(6)]
methods = lambda order:\
          [ ### "[spaceVersion,scheme]"
            # [{"version":"continuousSimplex"},"h1"],
            [ {"version":"continuous", "basisChoice":basisChoice},"vem"],
            # [ {"version":"non-conforming", "basisChoice":basisChoice},"vem"],
            # ["dg","bbdg"],
            # [ {"version":[0,order-2,order-3], "basisChoice":basisChoice},"vem"],
            # [ {"version":[0,order-2,order-1], "basisChoice":basisChoice},"vem"],
   ]
# <markdowncell>
# Now we define the model starting with the exact solution:

# <codecell>
uflSpace = dune.ufl.Space(2, dimRange=1)
x = SpatialCoordinate(uflSpace)
exact = as_vector( [x[0]*x[1] * cos(pi*x[0]*x[1])] )
exact = exact*10*((x[0]-1)*(x[0]+0.)*(x[1]-1)*(x[1]+0.))

# next the bilinear form
u = TrialFunction(uflSpace)
v = TestFunction(uflSpace)
a   = (diffCoeff*inner(grad(u),grad(v)) + massCoeff*dot(u,v) ) * dx
b   = (-diffCoeff*div(grad(exact[0])) + massCoeff*exact[0] ) * v[0] * dx
dbc = [dune.ufl.DirichletBC(uflSpace, exact, i+1) for i in range(4)]

# <markdowncell>
# Now we define a grid build up of voronoi cells around $50$ random points
# We now define a function to compute the solution and the $L^2,H^1$ error
# given a grid and a space

# <codecell>
def compute(grid, order, spaceArgs, schemeName):
    print("setting up space...",flush=True)
    space = generalSpace(polyGrid, order=order,
                         **spaceArgs,
                         dimRange=1, storage="fem")
    print("...done",flush=True)
    # solve the pde
    parameters = {"newton.linear.tolerance": 1e-12,
                  "newton.linear.preconditioning.method": "jacobi",
                  "penalty": diffCoeff*8*space.order**2,  # for the bbdg scheme
                  "newton.linear.maxiterations": 10000,
                  "newton.linear.verbose": False,
                  "newton.verbose": False
                  }
    if schemeName == "vem":
        scheme = create.scheme(schemeName, [a==b, *dbc], space,
                       solver="cg", # ("suitesparse","umfpack"),
                       gradStabilization=diffCoeff,
                       massStabilization=massCoeff,
                       parameters=parameters)
    else:
        scheme = create.scheme(schemeName, [a==b, *dbc], space,
                       solver="cg", # ("suitesparse","umfpack"),
                       parameters=parameters)
    df = space.interpolate([0],name="solution")

    if True:
        from dune.fem.operator import linear as linearOperator
        A = linearOperator(scheme).as_numpy.tocsc()
        # import numpy as np
        # Ad = A.todense()
        # print( "symmetry:", np.allclose(Ad, Ad.T, rtol=1e-7, atol=1e-9),flush=True )
        # l = np.linalg.eigvals(Ad).real
        # l.sort()
        # conditioning_dense = l[-1]/l[0]
        # print("conditioning dense", conditioning_dense)
        from scipy.sparse.linalg import eigs
        ll = eigs(A, k=1, which='LR', return_eigenvectors=False).real
        ll.sort()
        try:
            ls = eigs(A, k=1, which='LR', sigma=0,
                    return_eigenvectors=False,maxiter=20000,tol=1e-8).real
        except:
            print(A.todense())
            ls = eigs(A, k=1, which='SR',
                    return_eigenvectors=False,maxiter=20000,tol=1e-8).real
            print(ls)
        ls.sort()
        conditioning = ll[-1]/ls[0]
        print( "eigenvalues:", ls,ll,
               "conditioning:", conditioning,
               # "cond_dense:", conditioning_dense,
               flush=True )
    else:
        conditioning = -1

    info = {"linear_iterations":-1}
    df.interpolate(exact)
    # info = scheme.solve(target=df)

    # compute the error
    edf = exact-df
    errors = [ math.sqrt(e) for e in
               integrate(grid, [inner(edf,edf),inner(grad(edf),grad(edf))], order=8) ]

    info["conditioning"] = conditioning

    return df, errors, info

# <markdowncell>
# Finally we iterate over the requested methods and solve the problems

# <codecell>
results = []
for N in resolutions:
    constructor = cartesianDomain([0,0],[1,1],[N,N])
    # constructor = voronoiCells(constructor,N*N,"voronoiseeds",True)
    polyGrid = create.grid("agglomerate", constructor, cubes=False )
    for order in orders:
        # fig = pyplot.figure(figsize=(10*len(methods(order)),10))
        # figPos = 100+10*len(methods(order))+1
        for i,m in enumerate(methods(order)):
            print("Calculating: (",N,order,"):",i,m,flush=True)
            dfs,errors,info = compute(polyGrid, order,m[0],m[1])
            print("method:",m[0]," order: ",order," cells: ",N*N," simplex: ", polyGrid.size(0),
                  "size: ",dfs.size, "L^2: ", errors[0], "H^1: ", errors[1],
                  info["linear_iterations"],
                  info["conditioning"],
                  flush=True)
            print("Done",flush=True)
            results += [ [N,order,m,polyGrid.size(0),dfs.size,errors,info] ]
            # dfs.plot(figure=(fig,figPos+i),gridLines="black", colorbar="horizontal")
            del dfs
            print("collected: ",gc.collect(),flush=True)
        # pyplot.show()
        pickle.dump(results,open('testelliptic.dump','wb'))



import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.vem import voronoiCells
from dune.fem.operator import linear as linearOperator
from scipy.sparse.linalg import spsolve
import numpy

import ufl.algorithms
from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": -1})

maxLevel = 3

def interpolate():
    return False

def getParameters():
    ln, lm, Lx, Ly = 1,0, 1,1
    return ln, lm, Lx, Ly

def runTest(exact, spaceConstructor, get_df, N0=33):
    results = []
    for level in range(1,maxLevel):
        ln, lm, Lx, Ly = getParameters()
        # set up grid for testing
        N = 2**(level)
        # 23*N*N works, 19*N*N fails
        grid = dune.vem.polyGrid(
          dune.vem.voronoiCells([[0,0],[Lx,Ly]], N0*N*N, lloyd=250, load="voronoiseeds")
        #   cartesianDomain([0.,0.],[Lx,Ly],[N,N]), cubes=False
        #   cartesianDomain([0.,0.],[Lx,Ly],[2*N,2*N]), cubes=True
        )

        # get dimension of range
        dimRange = exact.ufl_shape[0]
        print('dim Range:', dimRange)

        res = []

        # construct space to use using spaceConstructor passed in
        space = spaceConstructor(grid,dimRange)

        err = get_df(space,exact)
        errors  = [ math.sqrt(e) for e in integrate(grid, err, order=8) ]
        length = len(errors)

        res += [ [[grid.hierarchicalGrid.agglomerate.size,space.size,*space.diameters()],*errors] ]
        results += [res]

    return calculateEOC(results,length)

def calculateEOC(results,length):
    eoc = length*[-1]

    for level in range(len(results)):
        print(*results[level][0],*eoc)
        if level<len(results)-1:
            eoc = [math.log(results[level+1][0][1+j]/results[level][0][1+j])/
                   math.log(results[level+1][0][0][3]/results[level][0][0][3])
                   for j in range(len(eoc))]

    return eoc

def checkEOC(eoc, expected_eoc):
    ret = 0
    i = 0
    for k in expected_eoc:
        # assert(0.8*k <= eoc[i]), "eoc out of expected range"
        if 0.8*k > eoc[i]:
            print(f"ERROR: {0.8*k} > {eoc[i]}")
            ret += 1
        if eoc[i] > 1.2*k:
            print("WARNING: the eoc seems too large (",eoc[i],"expected",k)
            ret += 1
        i += 1
    return ret



from concurrent.futures import ThreadPoolExecutor
class ThreadedGenerator:
    def __init__(self, maxWorkers):
        self.maxWorkers = maxWorkers
        self.builders = []
    def add(self, builder, *args, **kwargs):
        self.builders += [(builder,args,kwargs)]
    def clear(self):
        self.builders = []
    def execute(self):
        with ThreadPoolExecutor(max_workers = 4) as executor:
            thread = executor.map(ThreadedGenerator._gen, self.builders)
        self.clear()
        return [m for m in thread]
    @staticmethod
    def _gen(arg):
        builder = arg[0]
        args = arg[1]
        kwargs = arg[2]
        if builder:
            ret = builder(*args,**kwargs)
        else:
            ret = None
        # print("Building", builder, ret, flush=True)
        return ret

def main():
    def s(x,y):
        return x+y
    def m(x,y):
        return x*y
    def axpy(a,x,*,y):
        return y+a*x
    g = ThreadedGenerator(4)
    g.add(s,x=1,y=2)
    g.add(m,2,2)
    g.add(None) # this is needed to allow for a conditional module generation
    g.add(axpy,3,2,y=2)
    m = g.execute()
    print(m)

if __name__ == "__main__":
    main()



# from dune.vem._vem import *
from dune.vem.vem import *
from .voronoi import voronoiCells
# from dune.vem.voronoi import voronoiCells
registry = dict()
registry["space"] = {
        "bbdg"   : bbdgSpace,
        "vem"    : vemSpace
    }
registry["scheme"] = {
         "vem"    : vemScheme,
         "bbdg"   : bbdgScheme
    }
registry["grid"] = {
        "agglomerate" : polyGrid
   }



from ufl import *
from dune.ufl import Constant, DirichletBC
from dune.vem import polyGrid, voronoiCells, vemSpace, vemScheme
import dune.fem

dune.fem.threading.use = 8   # enable multithreading
conforming = True            # either conforming and non-conforming
order = 4                    # ... space of order 4
N     = 50                   # number of Voronoi cells
tau   = Constant(5e-6,"dt")  # time step size
saveTimes = [1e-4]           # add further time steps to extend simulation
# saveTimes += [4e-4,7e-4]

# moments for conforming and non-conforming C^1 spaces
if conforming:
    testSpaces=[[0,0],[order-4,order-3],order-4]
    name="willmore"+str(order)+"C"
else:
    testSpaces=[0,[order-3,order-2],order-4]
    name="willmore"+str(order)+"NC"

# setup grid and space
gridView = polyGrid( voronoiCells([[0,0],[1,1]],N,lloyd=200) )
space    = vemSpace(gridView, order=order, testSpaces=testSpaces)

# time, space coordinate, initial conditions and solution function
x   = SpatialCoordinate(space)
initial = sin(2*pi*x[0])**2*sin(2*pi*x[1])**2
df  = space.interpolate(initial, name="solution") # main solution

# Wilmore functional W(psi) and variational form
psi = Coefficient(space)
Q = lambda p: 1+inner(p,p)
E = lambda p: 1/Q(p)**0.25 * ( Identity(2) - outer(p,p)/Q(p) )
W = 1/2*inner( E(grad(psi)),grad(grad(psi)) )**2 * dx
phi = TestFunction(space)
a   = derivative(W,psi,phi)

# third order two stage time stepping (Gauss radau collocation)
# .. space with dimension 2 (df=U^n, U[0]=intermediate, U[1]=U^{n+1})
rkSpace = vemSpace(gridView, order=order,
                   dimRange=2,testSpaces=testSpaces)
stages  = rkSpace.interpolate([initial,initial], name="rk")
U       = TrialFunction(rkSpace)
V       = TestFunction(rkSpace)
dfs     = [  df,   U[0],  U[1]  ]
factors = [ [-2,  3./2., 1./2.] ,
            [ 2, -9./2., 5./2.] ]
dtForm  = lambda w,u,v: inner(w/Q(u),v)*dx
rkForm  = sum(f*dtForm(u,U[i],V[i])
              for i in range(2) for u,f in zip(dfs,factors[i]))
rkForm += tau*sum(replace(a,{psi:U[i],phi:V[i]}) for i in range(2))

# set up actual Newton solver using a direct solver for the linear part
dbc = DirichletBC(rkSpace, rkSpace.dimRange*[0])
factor = 40 # used for stabilization, determined experimentally
scheme = vemScheme( [rkForm == 0, dbc],
            solver=("suitesparse","umfpack"),
            hessStabilization=[factor*tau,factor*tau],
            gradStabilization=[factor*tau,factor*tau],
            massStabilization=[factor,factor],
            parameters={"newton.tolerance": 1e-6,
                        "newton.lineSearch": "simple"})

# time loop
t = 0

vtk = gridView.sequencedVTK(name, pointdata=[df], subsampling=2)
vtk()
saveNext = saveTimes.pop(0)
while True:
    info = scheme.solve(target=stages)
    # the second stage of the RK method is the approximation at t^{n+1}
    df.interpolate(stages[1])
    t += tau.value
    print("time",t,tau.value,info["iterations"],flush=True)
    # increase time step if Newton iterations drop to 5 or lower
    if info["iterations"]<=5: tau.value *= 1.25
    if t > saveNext:
        print("    saving....",flush=True)
        vtk()
        if len(saveTimes) == 0: break
        saveNext = saveTimes.pop(0)



from ufl import *
import dune.ufl

from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from script import runTest, checkEOC, interpolate
from interpolate import interpolate_secondorder

dimR = 3

def hk(space, exact):
    x = SpatialCoordinate(space)
    massCoeff = 1+sin(dot(x,x))
    diffCoeff = 1-0.9*cos(dot(x,x))
    assert(dimR == exact.ufl_shape[0]), "dim range not compatible"
#    dimR = exact.ufl_shape[0]
    u = TrialFunction(space)
    v = TestFunction(space)

    a = (diffCoeff*inner(grad(u),grad(v)) + massCoeff*dot(u,v) ) * dx
    b = sum( [ ( -div(diffCoeff*grad(exact[i])) + massCoeff*exact[i] ) * v[i]
             for i in range(dimR) ] ) * dx
    dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]

    df = space.interpolate(dimR*[0],name="solution")
    scheme = dune.vem.vemScheme(
                  [a==b, *dbc], space, solver="cg",
                  gradStabilization=diffCoeff,
                  massStabilization=massCoeff)
    info = scheme.solve(target=df)

    edf = exact-df
    err = [inner(edf,edf),
           inner(grad(edf),grad(edf))]

    return err

def runTesthk(testSpaces, order, vectorSpace, reduced):
    x = SpatialCoordinate(triangle)

    exact = as_vector( dimR*[x[0]*x[1] * cos(pi*x[0]*x[1])] )
    spaceConstructor = lambda grid, r: dune.vem.vemSpace( grid,
                                                          order=order,
                                                          dimRange=r,
                                                          testSpaces=testSpaces,
                                                          vectorSpace=vectorSpace,
                                                          reduced=reduced)
    expected_eoc = [order+1,order]
    if (interpolate()):
       eoc = runTest(exact, spaceConstructor, interpolate_secondorder)
    else:
       eoc = runTest(exact, spaceConstructor, hk)

    return eoc, expected_eoc

def main():
    # test elliptic with conforming and nonconforming second order VEM space
    orders = [1,3]
    for order in orders:
       print("order: ", order)

       with ProcessPoolExecutor(max_workers=4) as executor:
           # C0 non conforming VEM
           C0NCtestSpaces = [-1,order-1,order-2]
           print("C0 non conforming test spaces: ", C0NCtestSpaces)
           C0NCFuture = executor.submit( runTesthk, C0NCtestSpaces, order, False, False )

           # C0 conforming VEM
           C0testSpaces = [0,order-2,order-2]
           print("C0 conforming test spaces: ", C0testSpaces)
           C0Future = executor.submit( runTesthk, C0testSpaces, order, False, False )

           # C0 conforming VEM with vectorSpace=true
           print("C0 conforming test spaces with vectorSpace=True")
           C0vecFuture = executor.submit( runTesthk, C0testSpaces, order, True, False )

           # C0 serendipity VEM
           C0serendipitytestSpaces = [0,order-2,order-3]
           print("C0 serendipity test spaces: ", C0serendipitytestSpaces)
           C0serFuture = executor.submit( runTesthk, C0serendipitytestSpaces, order, False, False )


       # C0 non conforming VEM
       print("EOCs for C0 non conforming test spaces: ")
       eoc, expected_eoc = C0NCFuture.result()
       checkEOC(eoc, expected_eoc)

       # C0 conforming VEM
       print("EOCs for C0 conforming test spaces: ")
       eoc, expected_eoc = C0Future.result()
       checkEOC(eoc, expected_eoc)

       # C0 conforming VEM with vectorSpace=true
       print("EOCs for C0 conforming test spaces with vectorSpace=True")
       eoc, expected_eoc = C0vecFuture.result()
       checkEOC(eoc, expected_eoc)

       # C0 serendipity VEM
       print("EOCs for C0 serendipity test spaces: ")
       eoc, expected_eoc = C0serFuture.result()
       checkEOC(eoc, expected_eoc)

main()



import dune.vem
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math
from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.vem import voronoiCells
from dune.fem.operator import linear as linearOperator
from scipy.sparse.linalg import spsolve

from ufl import *
import dune.ufl

maxLevel     = 5
order        = 3

dune.fem.parameter.append({"fem.verboserank": 0})

methods = [ ### "[space,scheme,spaceKwrags]"
            ["vem","vem",{"testSpaces":[ [0],  [order-3,order-2], [order-4] ] }, "C1-non-conforming"],
            # ["vem","vem",{"testSpaces":[ [0],  [order-2,order-2], [order-2] ] }, "C1C0-conforming"],
            # ["vem","vem",{"testSpaces":[ [-1], [order-1,-1], [order-2] ] },   "C0-non-conforming"],
   ]
parameters = {"newton.linear.tolerance": 1e-8,
              "newton.tolerance": 5e-6,
              "newton.linear.preconditioning.method": "jacobi",
              "newton.lineSearch": "simple",
              "penalty": 40,  # for the bbdg scheme
              "newton.linear.verbose": False,
              "newton.verbose": True
              }


def compute(grid, s, schemeName):
    dt = 4e-4
    space = create.space(s[0], polyGrid, order=order, **s[1])
    t   = dune.ufl.Constant(0,"time")

    x   = SpatialCoordinate(space)
    initial = sin(2*pi*x[0])**2*sin(2*pi*x[1])**2

    # Wilmore functional W(psi)
    psi = Coefficient(space)
    Q = lambda p: 1+inner(p,p)
    E = lambda p: 1/Q(p)**0.25 * ( Identity(2) - outer(p,p)/Q(p) )
    Wint = 1/2*inner( E(grad(psi)),grad(grad(psi)) )**2
    W = Wint * dx

    # variation bilinear form
    phi = TestFunction(space)
    a   = derivative(W,psi,phi)

    # third order time stepping (Gauss radau collocation)
    tau = dune.ufl.Constant(0,"dt")
    df  = discreteFunction(space, name="solution") # main solution
    # space for time stepping (df=U^n, U[0]=intermediate, U[1]=U^{n+1}
    rkSpace = create.space(s[0], polyGrid, order=order, dimRange=2, **s[1])
    U       = TrialFunction(rkSpace)
    V       = TestFunction(rkSpace)
    dfs     = [  df,   U[0],  U[1]  ]
    factors = [ [-2,  3./2., 1./2.] ,
                [ 2, -9./2., 5./2.] ]
    dtForm = lambda w,u,v: inner(w/Q(u),v)*dx
    rkForm  = sum(f*dtForm(u,U[i],V[i]) for i in range(2) for u,f in zip(dfs,factors[i]))
    rkForm += tau*sum(replace(a,{psi:U[i],phi:V[i]}) for i in range(2))

    dbc = [dune.ufl.DirichletBC(rkSpace, rkSpace.dimRange*[0], i+1) for i in range(4)]
    biLaplaceCoeff = [dt,dt]
    diffCoeff      = [0,0]
    massCoeff      = [1,1] # 1/Q(U[i]) for i in range(2)]

    print("# solving: ",s, "size",space.size, "parameters:(", biLaplaceCoeff,",",diffCoeff,",",massCoeff,")")

    scheme = create.scheme(schemeName,
                        [rkForm == 0, *dbc], # space, # this gives is a compiler error
                        solver=("suitesparse","umfpack"),
                        hessStabilization=biLaplaceCoeff,
                        gradStabilization=diffCoeff,
                        massStabilization=massCoeff,
                        parameters=parameters)
    df.interpolate(initial)
    energy = integrate(grid,replace(Wint,{psi:df}),order=5)
    vtk = grid.sequencedVTK("voro", pointdata=[df],subsampling=2)
    t.value = 0
    tau.value = dt
    count = 0
    info = None
    # tmp = discreteFunction(rkSpace, name="rk")
    tmp = rkSpace.interpolate([initial,initial], name="rk")
    while t.value < 0.01:
        count += 1
        if count % 1 == 0:
            energy = integrate(grid,replace(Wint,{psi:df}),order=5)
            vtk()
        print("[",count,"]",t.value,energy,tau.value,info,flush=True)
        info = scheme.solve(target=tmp)
        t.value += tau
        df.interpolate(tmp[1])
    return df

for level in range(1,2): # maxLevel): # 2,3
    constructor = cartesianDomain([0,0],[1,1],[20*2**level,20*2**level])
    polyGrid = create.grid("agglomerate", constructor, cubes=False )
    # polyGrid = create.grid("agglomerate", voronoiCells(constructor,400*2**level*2**level,"voronoiseeds",load=True,show=False,lloyd=5) )
    @gridFunction(polyGrid, name="cells")
    def polygons(en,x):
        return polyGrid.hierarchicalGrid.agglomerate(en)
    # polygons.plot(colorbar="horizontal")
    for i,m in enumerate(methods):
        dfs = compute(polyGrid, [m[0], m[2]], m[1])
        print("[",level,"]","method:(",m[0],m[2],")",
              "Grid-Size: ",polyGrid.size(0), flush=True)



from ufl import *
import dune.ufl

from script import runTest, checkEOC, interpolate
from interpolate import interpolate_secondorder

parameters = {"newton.linear.tolerance": 1e-8,
              "newton.tolerance": 5e-6,
              "newton.lineSearch": "simple",
              "newton.linear.preconditioner": "ssor",
              "newton.linear.verbose": False,
              "newton.verbose": True
              }

mass = dune.ufl.Constant(1, "mu")

def divfree(space, exact):
    u = TrialFunction(space)
    v = TestFunction(space)

    a = (inner(grad(u),grad(v)) + mass*dot(u,v) ) * dx

    b = dot( -div(grad(exact))  + mass*exact, v) * dx
    dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]

    df = space.interpolate(exact,name="solution")
    scheme = dune.vem.vemScheme(
                  [a==b, *dbc], space,
                #   solver=("suitesparse","umfpack"),
                  parameters=parameters,
                  gradStabilization=1,
                  massStabilization=mass)
    info = scheme.solve(target=df)

    edf = exact-df
    err = [inner(edf,edf),
           inner(grad(edf),grad(edf))]

    return err

def runTestDivFree(order):
    x = SpatialCoordinate(triangle)

    p = (order+1)//2
    exact = as_vector([-x[1]+x[0]**(p+1)*x[1]**p,
                        x[0]-x[1]**(p+1)*x[0]**p])

    spaceConstructor = lambda grid, r: dune.vem.divFreeSpace( grid,
                                                              order=order )
    expected_eoc = [order+1]
    if (interpolate()):
      eoc = runTest(exact, spaceConstructor, interpolate_secondorder)
    else:
      eoc = runTest(exact, spaceConstructor, divfree)

    return eoc, expected_eoc

def main():
    orders = [2,3]
    for order in orders:
        print("order: ", order)

        eoc, expected_eoc = runTestDivFree( order )
        checkEOC(eoc, expected_eoc)

main()



from dune import create
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem.function import integrate, discreteFunction
from dune.fem import parameter
from dune.fem.operator import linear as linearOperator

import ufl.algorithms
from ufl import *
import dune.ufl

from script import runTest, checkEOC, interpolate
from interpolate import interpolate_secondorder

dimR = 3

parameters = {"newton.linear.tolerance": 1e-12,
              "newton.linear.preconditioning.method": "jacobi",
              "penalty": 40,  # for the bbdg scheme
              "newton.linear.verbose": False,
              "newton.verbose": True
              }

x = SpatialCoordinate(triangle)

def hk(space, exact):
    u      = TrialFunction(space)
    v      = TestFunction(space)
    normal = FacetNormal(space)

    diffCoeff = (0.1+dot(x,x))
    massCoeff = (2+dot(x,x))

    a = (diffCoeff*inner(grad(u),grad(v)) + massCoeff*dot(u,v) ) * dx
    b = sum( [ ( -div(diffCoeff*grad(exact[i])) + massCoeff*exact[i] ) * v[i]
             for i in range(dimR) ] ) * dx +\
        sum( [ diffCoeff*( dot(grad(exact[i]),normal) ) * v[i]
             for i in range(dimR) ] ) * ds
    a += dot(u,v) * ds
    b += dot(exact,v) * ds

    scheme = dune.vem.vemScheme([a==b], space,
                            solver="cg",
                            gradStabilization=diffCoeff,
                            massStabilization=massCoeff,
                            parameters=parameters)

    df = discreteFunction(space, name="solution")
    info = scheme.solve(target=df)

    edf = exact-df
    err = [inner(edf,edf),
            inner(grad(edf),grad(edf))]

    return err

def runTesthk(testSpaces, order, vectorSpace, reduced):
    exact = as_vector( dimR*[x[0]**dimR*x[1] * cos(pi*x[0]*x[1]**dimR)] )
    spaceConstructor = lambda grid, r: dune.vem.vemSpace( grid,
                                                          order=order,
                                                          dimRange=r,
                                                          testSpaces=testSpaces,
                                                          vectorSpace=vectorSpace,
                                                          reduced=reduced)

    expected_eoc = [order+1, order]
    if (interpolate()):
        eoc = runTest(exact, spaceConstructor, interpolate_secondorder)
    else:
        eoc = runTest(exact, spaceConstructor, hk)

    return eoc, expected_eoc

def main():
    orders = [1,3]
    for order in orders:
        print("order: ", order)
        # C0 conforming VEM
        C0testSpaces = [0,order-2,order-2]
        print("C0 conforming test spaces: ", C0testSpaces)

        # vectorSpace=False, reduced=True
        eoc, expected_eoc = runTesthk( C0testSpaces, order, vectorSpace=False, reduced=True )
        checkEOC(eoc, expected_eoc)

main()



import logging, io
logger = logging.getLogger(__name__)
import numpy as np
from matplotlib import pyplot as plt

from ufl.equation import Equation
from ufl import Form
from dune.generator import Constructor, Method, algorithm
import dune
import dune.fem

def stabilization(spc, hessStabilization=None, gradStabilization=None, massStabilization=None):
    if hessStabilization is None and gradStabilization is None and massStabilization is None:
        gradStabilization = 1.
    if hessStabilization is None:
        hessStabilization = 0.
    else:
        hessStabilization = float(hessStabilization)
    if gradStabilization is None:
        gradStabilization = 0.
    else:
        gradStabilization = float(gradStabilization)
    if massStabilization is None:
        massStabilization = 0.
    else:
        massStabilization = float(massStabilization)
    if spc._stab is None:
        stabMatrix = dune.fem.operator.linear([spc,spc])
        spc._stab = algorithm.load("Dune::Vem::stabilization",
           io.StringIO("#include <dune/vem/operator/stabmatrix.hh>"), stabMatrix,
                       hessStabilization,gradStabilization,massStabilization)
    else:
        stabMatrix = dune.fem.operator.linear([spc,spc])
    spc._stab(stabMatrix, hessStabilization, gradStabilization, massStabilization)
    return stabMatrix


def bbdgSpace(view, order=1, scalar=False, dimRange=None, field="double",
             rotatedBB=True,
             storage="numpy"):
    """create a discontinous galerkin space over an agglomerated grid

    Args:
        view: the underlying grid view
        order: polynomial order of the finite element functions
        dimRange: dimension of the range space
        field: field of the range space
        storage: underlying linear algebra backend

    Returns:
        Space: the constructed Space
    """

    from dune.fem.space import module, addStorage
    if dimRange is None:
        dimRange = 1
        scalar = True

    if dimRange > 1 and scalar:
        raise KeyError(\
                "trying to set up a scalar space with dimRange = " +\
                str(dimRange) + ">1")
    if dimRange < 1:
        raise KeyError(\
            "Parameter error in BBDGSpace with "+
            "dimRange=" + str(dimRange) + ": " +\
            "dimRange has to be greater or equal to 1")
    if order < 0:
        raise KeyError(\
            "Parameter error in BBDGSpace with "+
            "order=" + str(order) + ": " +\
            "order has to be greater or equal to 0")
    if field == "complex":
        field = "std::complex<double>"

    agglomerate = view.hierarchicalGrid.agglomerate

    includes = [ "dune/vem/agglomeration/dgspace.hh" ] + view.cppIncludes
    dimw = view.dimWorld
    viewType = view.cppTypeName

    gridPartName = "Dune::FemPy::GridPart< " + view.cppTypeName + " >"
    typeName = "Dune::Vem::AgglomerationDGSpace< " +\
      "Dune::Fem::FunctionSpace< double, " + field + ", " + str(dimw) + ", " + str(dimRange) + " >, " +\
      gridPartName + ", " + str(order) + " >"

    constructor = Constructor(
                   ['pybind11::handle gridView',
                    'const pybind11::object agglomerate',
                    'bool rotatedBB'],
                   ['typedef Dune::Vem::Agglomeration<' + gridPartName + '> AggloType;',
                    'AggloType *agglo = nullptr;',
                    '// check storage',
                    'pybind11::function retrieve = pybind11::module::import("dune.vem.vem").attr("retrieveAgglo_");',
                    'pybind11::function insert   = pybind11::module::import("dune.vem.vem").attr("insertAgglo_");',
                    'pybind11::function remove   = pybind11::module::import("dune.vem.vem").attr("removeAgglo_");',
                    'pybind11::handle aggloExists = retrieve(gridView);',
                    'if (! pybind11::isinstance<pybind11::none>(aggloExists) ) {',
                    '  agglo = static_cast<AggloType*>(aggloExists.cast<void*>());',
                    '}',
                    'else {',
                    '  agglo = new AggloType(',
                    '           Dune::FemPy::gridPart<' + viewType + '>(gridView),rotatedBB,',
                    '      [agglomerate](const auto& idx)',
                    '      { return agglomerate(idx).template cast<unsigned int>(); } );',
                    '  insert(gridView,(void*)agglo);',
                    '  pybind11::cpp_function aggloCleanup(',
                    '        [agglo,remove](pybind11::handle weakref) {',
                    '        remove((void*)agglo);',
                    '        delete agglo;',
                    '        weakref.dec_ref();',
                    '    });',
                    '  (void) pybind11::weakref(gridView, aggloCleanup).release();',
                    '}',
                    'auto obj = new DuneType( *agglo );',
                    'return obj;'],
                   ['"gridView"_a', '"agglomerate"_a', '"rotatedBB"_a',
                    'pybind11::keep_alive< 1, 2 >()'] )


    # is defined on the index set which is not used for the dg space
    # diameterMethod = Method('diameters',
    #    '''[]( DuneType &self ) { return self.blockMapper().indexSet().diameters(); }''' )
    updateMethod = Method('update',
       '''[]( DuneType &self ) { self.update(); }''' )

    spc = module(field, includes, typeName, constructor, updateMethod, scalar=scalar, storage=storage,
             ctorArgs=[view, agglomerate,rotatedBB])
    addStorage(spc, storage)
    return spc.as_ufl()

from dune.fem.scheme import dg
def bbdgScheme(model, space=None, penalty=1, solver=None, parameters={}):
    if space == None:
        try:
            space = model.lhs.arguments()[0].ufl_function_space()
        except AttributeError:
            raise ValueError("no space provided and could not deduce from form provided")
    spaceType = space.cppTypeName
    penaltyClass = "Dune::Vem::BBDGPenalty<"+spaceType+">"
    return dg(model,space,penalty,solver,parameters,penaltyClass)
    # return galerkin(model,space,solver,parameters)

def vemSpace(view, order=1,
             testSpaces=None, scalar=False,
             dimRange=None, conforming=True, field="double",
             storage="numpy",
             basisChoice=2, rotatedBB=True,
             edgeInterpolation=False,
             vectorSpace=False, reduced=False,
             basisField="double", computeField="long double",
             gradProjectionWithIByParts=False
             ):
    """create a virtual element space over an agglomerated grid

    Args:
        view: the underlying grid view
        order: polynomial order of the finite element functions
        dimRrange: dimension of the range space
        field: field of the range space
        storage: underlying linear algebra backend

    Returns:
        Space: the constructed Space
    """
    try:
        orderTuple = order
        order = orderTuple[0]
    except TypeError:
        orderTuple = [-1,-1,-1]

    if gradProjectionWithIByParts:
        useNewGradProjection = ["NEWGRADPROJECTION"]
    else:
        useNewGradProjection = []

    from dune.fem.space import module, addStorage
    if dimRange is None:
        dimRange = 1
        scalar = True

    if dimRange > 1 and scalar:
        raise KeyError(\
                "trying to set up a scalar space with dimRange = " +\
                str(dimRange) + ">1")
    if dimRange < 1:
        raise KeyError(\
            "Parameter error in vemSpace with "+
            "dimRange=" + str(dimRange) + ": " +\
            "dimRange has to be greater or equal to 1")
    if order < 0:
        raise KeyError(\
            "Parameter error in vemSpace with "+
            "order=" + str(order) + ": " +\
            "order has to be greater or equal to 0")
    if field == "complex":
        field = "std::complex<double>"

    if testSpaces is None:
        if conforming:
            testSpaces = [  [0], [order-2], [order-2] ]
        else:
            testSpaces = [ [-1], [order-1], [order-2] ]
    if testSpaces == "serendipity":
        etaMin = view.hierarchicalGrid.agglomerate.minEdgeNumber
        testSpaces = [  [0], [order-2], [max(-1,order-etaMin)] ]
        print("serendipity:",testSpaces,etaMin)

    for i,ts in enumerate(testSpaces):
        if type(ts) == int:
            testSpaces[i] = [ts]

    try:
        agglomerate = view.hierarchicalGrid.agglomerate
    except AttributeError:
        # if view does not have agglomerate then we create a trivial one
        # this only works for quads or triangles
        view.hierarchicalGrid.agglomerate = TrivialAgglomerate(view)
        agglomerate = view.hierarchicalGrid.agglomerate

    includes = [ "dune/fem/gridpart/common/gridpart.hh", "dune/vem/space/hk.hh" ] + view.cppIncludes
    dimw = view.dimWorld
    viewType = view.cppTypeName

    gridPartName = "Dune::FemPy::GridPart< " + view.cppTypeName + " >"
    typeName = "Dune::Vem::AgglomerationVEMSpace< " +\
      "Dune::Fem::FunctionSpace< double, " + field + ", " + str(dimw) + ", " + str(dimRange) + " >, " +\
      gridPartName +\
      (", true " if vectorSpace else ", false ") +\
      (", true " if reduced else ", false ") +\
      ", " + basisField + ", " + computeField +\
      " >"
    constructor = Constructor(
                   ['pybind11::handle gridView',
                    'const pybind11::object agglomerate',
                    'unsigned int order',
                    'const std::array<int,3> orderTuple',
                    'const std::array<std::vector<int>,'+str(view.dimension+1)+'> &testSpaces',
                    'int basisChoice','bool edgeInterpolation','bool rotatedBB'],
                   ['typedef Dune::Vem::Agglomeration<' + gridPartName + '> AggloType;',
                    'AggloType *agglo = nullptr;',
                    '// check storage',
                    'pybind11::function retrieve = pybind11::module::import("dune.vem.vem").attr("retrieveAgglo_");',
                    'pybind11::function insert   = pybind11::module::import("dune.vem.vem").attr("insertAgglo_");',
                    'pybind11::function remove   = pybind11::module::import("dune.vem.vem").attr("removeAgglo_");',
                    'pybind11::handle aggloExists = retrieve(gridView);',
                    'if (! pybind11::isinstance<pybind11::none>(aggloExists) ) {',
                    '  // std::cout << "retrieve agglo\\n";',
                    '  agglo = static_cast<AggloType*>(aggloExists.cast<void*>());',
                    '}',
                    'else {',
                    '  agglo = new AggloType(',
                    '           Dune::FemPy::gridPart<' + viewType + '>(gridView),rotatedBB,',
                    '      [agglomerate](const auto& idx)',
                    '      { return agglomerate(idx).template cast<unsigned int>(); } );',
                    '  insert(gridView,(void*)agglo);',
                    '  pybind11::cpp_function aggloCleanup(',
                    '        [agglo,remove](pybind11::handle weakref) {',
                    '        // std::cout << "remove agglo\\n";',
                    '        remove((void*)agglo);',
                    '        delete agglo;',
                    '        weakref.dec_ref();',
                    '    });',
                    '  (void) pybind11::weakref(gridView, aggloCleanup).release();',
                    '  // std::cout << "new agglo\\n";',
                    '}',
                    'auto obj = new DuneType(*agglo, order, orderTuple, testSpaces, basisChoice, edgeInterpolation );',
                    'return obj;'],
                   ['"gridView"_a', '"agglomerate"_a',
                    '"order"_a',
                    '"orderTuple"_a',
                    '"testSpaces"_a',
                    '"basisChoice"_a', '"edgeInterpolation"_a', '"rotatedBB"_a',
                    'pybind11::keep_alive< 1, 2 >()'] )
    diameterMethod = Method('diameters',
       '''[]( DuneType &self ) { return self.blockMapper().indexSet().diameters(); }''' )
    updateMethod = Method('update',
       '''[]( DuneType &self ) { self.update(); }''' )

    from dune.fem.space import _defaultGenerator
    class NewGenerator:
        def __init__(self):
            pass
        def load(self,*args,**kwargs):
            kwargs["defines"] = kwargs.get("defines",[]) + useNewGradProjection
            return _defaultGenerator.load(*args,**kwargs)

    spc = module(field, includes, typeName, constructor, diameterMethod, updateMethod,
                generator = NewGenerator(),
                scalar=scalar, storage=storage,
                ctorArgs=[view, agglomerate, order, orderTuple, testSpaces, basisChoice, edgeInterpolation, rotatedBB])
    addStorage(spc, storage)

    spc._stab = None
    spc.__class__.stabilization = stabilization

    return spc.as_ufl()

def space(view, order=1, dimRange=None,
          testSpaces=None,
          conforming=True,
          version=None,
          scalar=False,
          basisChoice=2,
          field="double", storage="numpy"):
    '''
        version is tuple,list for a general vem space (=tests[aces)
        version == "dg" (bbdg)
        version == "dgSimplex" (dg on triangulation)
        version == "continuous" (vem)
        version == "continuousSimplex" (lagrange)
        version == "non-conformin" (vem)
    '''
    if version is None and testSpaces is None:
        raise AttributeError("version and testSpaces parameters can not both be None (default)")
    if version is None: version = testSpaces
    if isinstance(version,tuple) or isinstance(version,list):
        return vemSpace(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage,
                        testSpaces=version, basisChoice=basisChoice)
    elif version == "dg":
        return bbdgSpace(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage)
    elif version == "dgSimplex":
        from dune.fem.space import onb
        return onb(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage)
    elif version == "continuous":
        return vemSpace(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage,
                        conforming=True, basisChoice=basisChoice)
    elif version == "continuousSimplex":
        from dune.fem.space import lagrange
        return lagrange(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage)
    elif version == "non-conforming":
        return vemSpace(view,order=order,scalar=scalar,dimRange=dimRange,field=field,storage=storage,
                        conforming=False, basisChoice=basisChoice)
    raise AttributeError("wrong version string provided:",version)

###############################

def curlFreeSpace(view, order=1,
             field="double", storage="numpy",
             basisChoice=2, rotatedBB=True):
    """create a virtual element space over an agglomerated grid

    Args:
        view: the underlying grid view
        order: polynomial order of the finite element functions
        field: field of the range space
        storage: underlying linear algebra backend

    Returns:
        Space: the constructed Space
    """

    from dune.fem.space import module, addStorage

    if order < 0:
        raise KeyError(\
            "Parameter error in curlFreeSpace with "+
            "order=" + str(order) + ": " +\
            "order has to be greater or equal to 0")

    agglomerate = view.hierarchicalGrid.agglomerate

    includes = [ "dune/fem/gridpart/common/gridpart.hh", "dune/vem/space/curlfree.hh" ] + view.cppIncludes
    dimw = view.dimWorld
    viewType = view.cppTypeName

    gridPartName = "Dune::FemPy::GridPart< " + view.cppTypeName + " >"
    typeName = "Dune::Vem::CurlFreeVEMSpace< " + gridPartName + " >"
    constructor = Constructor(
                   ['pybind11::handle gridView',
                    'const pybind11::object agglomerate',
                    'unsigned int order',
                    'int basisChoice','bool rotatedBB'],
                   ['typedef Dune::Vem::Agglomeration<' + gridPartName + '> AggloType;',
                    'AggloType *agglo = nullptr;',
                    '// check storage',
                    'pybind11::function retrieve = pybind11::module::import("dune.vem.vem").attr("retrieveAgglo_");',
                    'pybind11::function insert   = pybind11::module::import("dune.vem.vem").attr("insertAgglo_");',
                    'pybind11::function remove   = pybind11::module::import("dune.vem.vem").attr("removeAgglo_");',
                    'pybind11::handle aggloExists = retrieve(gridView);',
                    'if (! pybind11::isinstance<pybind11::none>(aggloExists) ) {',
                    '  // std::cout << "retrieve agglo\\n";',
                    '  agglo = static_cast<AggloType*>(aggloExists.cast<void*>());',
                    '}',
                    'else {',
                    '  agglo = new AggloType(',
                    '           Dune::FemPy::gridPart<' + viewType + '>(gridView),rotatedBB,',
                    '      [agglomerate](const auto& idx)',
                    '      { return agglomerate(idx).template cast<unsigned int>(); } );',
                    '  insert(gridView,(void*)agglo);',
                    '  pybind11::cpp_function aggloCleanup(',
                    '        [agglo,remove](pybind11::handle weakref) {',
                    '        // std::cout << "remove agglo\\n";',
                    '        remove((void*)agglo);',
                    '        delete agglo;',
                    '        weakref.dec_ref();',
                    '    });',
                    '  (void) pybind11::weakref(gridView, aggloCleanup).release();',
                    '  // std::cout << "new agglo\\n";',
                    '}',
                    'auto obj = new DuneType(*agglo, order, basisChoice );',
                    'return obj;'],
                   ['"gridView"_a', '"agglomerate"_a',
                    '"order"_a',
                    '"basisChoice"_a', '"rotatedBB"_a',
                    'pybind11::keep_alive< 1, 2 >()'] )
    diameterMethod = Method('diameters',
       '''[]( DuneType &self ) { return self.blockMapper().indexSet().diameters(); }''' )
    updateMethod = Method('update',
       '''[]( DuneType &self ) { self.update(); }''' )

    spc = module(field, includes, typeName, constructor, diameterMethod, updateMethod,
                storage=storage,
                ctorArgs=[view, agglomerate, order, basisChoice, rotatedBB])
    addStorage(spc, storage)
    spc._stab = None
    spc.stabilization = lambda: stabilization(spc)
    return spc.as_ufl()

#########################################################

def divFreeSpace(view, order=2, conforming=True,
             field="double", storage="numpy",
             basisChoice=2, edgeInterpolation=False, rotatedBB=True):
    """create a virtual element space over an agglomerated grid

    Args:
        view: the underlying grid view
        order: polynomial order of the finite element functions
        field: field of the range space
        storage: underlying linear algebra backend

    Returns:
        Space: the constructed Space
    """

    from dune.fem.space import module, addStorage

    if order < 2:
        raise KeyError(\
            "Parameter error in divFreeSpace with "+
            "order=" + str(order) + ": " +\
            "order has to be greater or equal to 2")

    agglomerate = view.hierarchicalGrid.agglomerate

    includes = [ "dune/fem/gridpart/common/gridpart.hh", "dune/vem/space/divfree.hh" ] + view.cppIncludes
    dimw = view.dimWorld
    viewType = view.cppTypeName

    gridPartName = "Dune::FemPy::GridPart< " + view.cppTypeName + " >"
    typeName = "Dune::Vem::DivFreeVEMSpace< " + gridPartName + " >"
    constructor = Constructor(
                   ['pybind11::handle gridView',
                    'const pybind11::object agglomerate',
                    'unsigned int order', 'bool conforming',
                    'int basisChoice','bool edgeInterpolation','bool rotatedBB'],
                   ['typedef Dune::Vem::Agglomeration<' + gridPartName + '> AggloType;',
                    'AggloType *agglo = nullptr;',
                    '// check storage',
                    'pybind11::function retrieve = pybind11::module::import("dune.vem.vem").attr("retrieveAgglo_");',
                    'pybind11::function insert   = pybind11::module::import("dune.vem.vem").attr("insertAgglo_");',
                    'pybind11::function remove   = pybind11::module::import("dune.vem.vem").attr("removeAgglo_");',
                    'pybind11::handle aggloExists = retrieve(gridView);',
                    'if (! pybind11::isinstance<pybind11::none>(aggloExists) ) {',
                    '  // std::cout << "retrieve agglo\\n";',
                    '  agglo = static_cast<AggloType*>(aggloExists.cast<void*>());',
                    '}',
                    'else {',
                    '  agglo = new AggloType(',
                    '           Dune::FemPy::gridPart<' + viewType + '>(gridView),rotatedBB,',
                    '      [agglomerate](const auto& idx)',
                    '      { return agglomerate(idx).template cast<unsigned int>(); } );',
                    '  insert(gridView,(void*)agglo);',
                    '  pybind11::cpp_function aggloCleanup(',
                    '        [agglo,remove](pybind11::handle weakref) {',
                    '        // std::cout << "remove agglo\\n";',
                    '        remove((void*)agglo);',
                    '        delete agglo;',
                    '        weakref.dec_ref();',
                    '    });',
                    '  (void) pybind11::weakref(gridView, aggloCleanup).release();',
                    '  // std::cout << "new agglo\\n";',
                    '}',
                    'auto obj = new DuneType(*agglo, order, conforming, basisChoice, edgeInterpolation );',
                    'return obj;'],
                   ['"gridView"_a', '"agglomerate"_a',
                    '"order"_a', '"conforming"_a',
                    '"basisChoice"_a', '"edgeInterpolation"_a','"rotatedBB"_a',
                    'pybind11::keep_alive< 1, 2 >()'] )
    diameterMethod = Method('diameters',
       '''[]( DuneType &self ) { return self.blockMapper().indexSet().diameters(); }''' )
    updateMethod = Method('update',
       '''[]( DuneType &self ) { self.update(); }''' )

    spc = module(field, includes, typeName, constructor, diameterMethod, updateMethod,
                storage=storage,
                ctorArgs=[view, agglomerate, order, conforming, basisChoice, edgeInterpolation, rotatedBB])
    addStorage(spc, storage)
    spc._stab = None
    spc.stabilization = lambda: stabilization(spc)
    return spc.as_ufl()

#########################################################
#########################################################

# from dune.fem.model import elliptic
from dune.fem.model import integrands
from dune.vem.patch import transform
from dune.fem.model import warnNewCartesianIds

def vemModel(view, equation, space,
        hessStabilization=None,gradStabilization=None, massStabilization=None,
        *args, **kwargs):
    warnNewCartesianIds(view,*args)
    return integrands(view, equation,
                      modelPatch=transform(space,hessStabilization,gradStabilization,massStabilization),
                      includes=["dune/vem/py/integrands.hh"],
                      *args, **kwargs)

def vemScheme(model, space=None, solver=None, parameters={},
              hessStabilization=None, gradStabilization=None, massStabilization=None,
              boundary="default"):
    """create a scheme for solving second order pdes with the virtual element method

    Args:

    Returns:
        Scheme: the constructed scheme
    """

    from dune.fem.scheme import module
    from dune.fem.scheme import femschemeModule

    assert boundary == "default" or boundary=="full" or boundary=="value" or boundary=="derivative" or boundary is None

    modelParam = None
    if isinstance(model, (list, tuple)):
        modelParam = model[1:]
        model = model[0]
    if isinstance(model,Equation):
        if space == None:
            try:
                space = model.lhs.arguments()[0].ufl_function_space()
            except AttributeError:
                raise ValueError("no space provided and could not deduce from form provided")
        else:
            try:
                eqSpace = model.lhs.arguments()[0].ufl_function_space()
                if not eqSpace.dimRange == space.dimRange:
                    raise ValueError("""range of space used for arguments in equation
                    and the range of space passed to the scheme have to match -
                    space argument to scheme can be 'None'""")
            except AttributeError:
                pass
        if modelParam:
            model = vemModel(space.gridView,model,space,hessStabilization,gradStabilization,massStabilization,*modelParam)
        else:
            model = vemModel(space.gridView,model,space,hessStabilization,gradStabilization,massStabilization)

    includes = [ "dune/vem/operator/vemelliptic.hh", "dune/vem/operator/diffusionmodel.hh" ]

    op = lambda linOp,model: "DifferentiableVEMEllipticOperator< " +\
                             ",".join([linOp,model]) + ">"

    if model.hasDirichletBoundary:
        if boundary == "default": boundary = "full"
        assert boundary is not None
        includes += [ "dune/fem/schemes/dirichletwrapper.hh",
                      "dune/vem/operator/vemdirichletconstraints.hh"]
        if boundary   == "full":       boundary = 3
        elif boundary == "value":      boundary = 1
        elif boundary == "derivative": boundary = 2
        constraints = lambda model: "Dune::VemDirichletConstraints< " +\
                    ",".join([model,space.cppTypeName,str(boundary)]) + " > "
        operator = lambda linOp,model: "DirichletWrapperOperator< " +\
                ",".join([op(linOp,model),constraints(model)]) + " >"
    elif boundary == 'derivative':
        includes += [ "dune/fem/schemes/dirichletwrapper.hh",
                      "dune/vem/operator/vemdirichletconstraints.hh"]
        constraints = lambda model: "Dune::VemDirichletConstraints< " +\
                    ",".join([model,space.cppTypeName,str(2)]) + " > "
        operator = lambda linOp,model: "DirichletWrapperOperator< " +\
                ",".join([op(linOp,model),constraints(model)]) + " >"
    else:
        assert boundary is None or boundary == "default"
        operator = op

    spaceType = space.cppTypeName
    includes += ["dune/vem/operator/diffusionmodel.hh"]
    valueType = 'std::tuple< typename ' + spaceType + '::RangeType, typename ' + spaceType + '::JacobianRangeType >'
    modelType = 'Dune::Fem::VirtualizedVemIntegrands< typename ' + spaceType + '::GridPartType, ' + model._domainValueType + ", " + model._rangeValueType+ ' >'

    return femschemeModule(space,model,includes,solver,operator,
            parameters=parameters,
            modelType=modelType
            )


def vemOperator(model, domainSpace=None, rangeSpace=None, boundary="default"):
    # from dune.fem.model._models import elliptic as makeElliptic
    from dune.fem.operator import load
    if rangeSpace is None:
        rangeSpace = domainSpace

    modelParam = None
    if isinstance(model, (list, tuple)):
        modelParam = model[1:]
        model = model[0]
    if isinstance(model,Form):
        model = model == 0
    if isinstance(model,Equation):
        if rangeSpace == None:
            try:
                rangeSpace = model.lhs.arguments()[0].ufl_function_space()
            except AttributeError:
                raise ValueError("no range space provided and could not deduce from form provided")
        if domainSpace == None:
            try:
                domainSpace = model.lhs.arguments()[1].ufl_function_space()
            except AttributeError:
                raise ValueError("no domain space provided and could not deduce from form provided")
        if modelParam:
            model = vemModel(domainSpace.gridView,model,domainSpace,None,None,None,*modelParam)
        else:
            model = vemModel(domainSpace.gridView,model,domainSpace,None,None,None)

    if not hasattr(rangeSpace,"interpolate"):
        raise ValueError("wrong range space")
    if not hasattr(domainSpace,"interpolate"):
        raise ValueError("wrong domain space")

    domainSpaceType = domainSpace.cppTypeName
    rangeSpaceType = rangeSpace.cppTypeName

    storage,  domainFunctionIncludes, domainFunctionType, _, _, dbackend = domainSpace.storage
    rstorage, rangeFunctionIncludes,  rangeFunctionType,  _, _, rbackend = rangeSpace.storage
    if not rstorage == storage:
        raise ValueError("storage for both spaces must be identical to construct operator")

    includes = ["dune/vem/operator/vemelliptic.hh",
                "dune/fempy/py/grid/gridpart.hh",
                "dune/fem/schemes/dirichletwrapper.hh",
                "dune/vem/operator/vemdirichletconstraints.hh"]
    includes += domainSpace.cppIncludes + domainFunctionIncludes
    includes += rangeSpace.cppIncludes + rangeFunctionIncludes
    includes += ["dune/vem/operator/diffusionmodel.hh", "dune/fempy/parameter.hh"]

    import dune.create as create
    linearOperator = create.discretefunction(storage)(domainSpace,rangeSpace)[3]

    modelType = "VEMDiffusionModel< " +\
          "typename " + domainSpaceType + "::GridPartType, " +\
          domainSpaceType + "::dimRange, " +\
          rangeSpaceType + "::dimRange, " +\
          "typename " + domainSpaceType + "::RangeFieldType >"
    typeName = "DifferentiableVEMEllipticOperator< " + linearOperator + ", " + modelType + ">"
    if model.hasDirichletBoundary:
        if boundary == "default": boundary = "full"
        assert boundary is not None
        includes += [ "dune/fem/schemes/dirichletwrapper.hh",
                      "dune/vem/operator/vemdirichletconstraints.hh"]
        if boundary   == "full":       boundary = 3
        elif boundary == "value":      boundary = 1
        elif boundary == "derivative": boundary = 2
        constraints = lambda model: "Dune::VemDirichletConstraints< " +\
                    ",".join([model,rangeSpace.cppTypeName,str(boundary)]) + " > "
        typeName = "DirichletWrapperOperator< " +\
                ",".join([typeName,constraints(modelType)]) + " >"

    constructor = Constructor(['const '+domainSpaceType+'& dSpace','const '+rangeSpaceType+' &rSpace', modelType + ' &model'],
                              ['return new ' + typeName + '( dSpace, rSpace, model );'],
                              ['pybind11::keep_alive< 1, 2 >()', 'pybind11::keep_alive< 1, 3 >()', 'pybind11::keep_alive< 1, 4 >()'])

    scheme = load(includes, typeName, constructor).Operator(domainSpace,rangeSpace, model)
    scheme.model = model
    return scheme

#################################################################

def aluGrid(constructor, dimgrid=None, dimworld=None, elementType=None, **parameters):
    if not dimgrid:
        dimgrid = getDimgrid(constructor)
    if dimworld is None:
        dimworld = dimgrid
    if elementType is None:
        elementType = parameters.pop("type")
    refinement = parameters["refinement"]
    if refinement == "conforming":
        refinement="Dune::conforming"
    elif refinement == "nonconforming":
        refinement="Dune::nonconforming"
    if not (2 <= dimworld and dimworld <= 3):
        raise KeyError("Parameter error in ALUGrid with dimworld=" + str(dimworld) + ": dimworld has to be either 2 or 3")
    if not (2 <= dimgrid and dimgrid <= dimworld):
        raise KeyError("Parameter error in ALUGrid with dimgrid=" + str(dimgrid) + ": dimgrid has to be either 2 or 3")
    if refinement=="Dune::conforming" and elementType=="Dune::cube":
        raise KeyError("Parameter error in ALUGrid with refinement=" + refinement + " and type=" + elementType + ": conforming refinement is only available with simplex element type")
    typeName = "Dune::ALUGrid< " + str(dimgrid) + ", " + str(dimworld) + ", " + elementType + ", " + refinement + " >"
    includes = ["dune/alugrid/grid.hh", "dune/alugrid/dgf.hh"]
    gridModule = module(includes, typeName) # , dynamicAttr=True)
    return gridModule.LeafGrid(gridModule.reader(constructor))
def aluSimplexGrid(constructor, dimgrid=2, dimworld=2):
    from dune.grid.grid_generator import module, getDimgrid
    typeName = "Dune::Vem::Grid<"+str(dimgrid)+","+str(dimworld)+">"
    includes = ["dune/vem/misc/grid.hh"]
    gridModule = module(includes, typeName) # , dynamicAttr=True)
    return gridModule.LeafGrid(gridModule.reader(constructor))
def aluCubeGrid(constructor, dimgrid=2, dimworld=2):
    from dune.grid.grid_generator import module, getDimgrid
    typeName = "Dune::Vem::CubeGrid<"+str(dimgrid)+","+str(dimworld)+">"
    includes = ["dune/vem/misc/grid.hh"]
    gridModule = module(includes, typeName) # , dynamicAttr=True)
    return gridModule.LeafGrid(gridModule.reader(constructor))
def yaspGrid(constructor, dimgrid=None, coordinates="equidistant", ctype="double"):
    from dune.generator import Constructor
    from dune.grid.grid_generator import module, getDimgrid
    from dune.typeregistry import generateTypeName

    if not dimgrid:
        dimgrid = getDimgrid(constructor)
    if coordinates == "equidistant":
        coordinates = "Dune::EquidistantOffsetCoordinates"
    elif coordinates == "tensorproduct":
        coordinates = "Dune::TensorProductCoordinates"

    if not dimgrid:
        dimgrid = getDimgrid(constructor)
    coordinates, _ = generateTypeName(coordinates, ctype, dimgrid)
    typeName, includes = generateTypeName("Dune::Vem::YGrid", dimgrid, coordinates)
    includes += ["dune/vem/misc/grid.hh", "dune/grid/io/file/dgfparser/dgfyasp.hh"]

    ctor = Constructor([
              'Dune::FieldVector<'+ctype+', '+str(dimgrid)+'> lowerleft',
              'Dune::FieldVector<'+ctype+', '+str(dimgrid)+'> upperright',
              'std::array<int, '+str(dimgrid)+'> elements',
              'std::array<bool, '+str(dimgrid)+'> periodic',
              'int overlap'],
              ['std::bitset<'+str(dimgrid)+'> periodic_;',
               'for (int i=0;i<'+str(dimgrid)+';++i) periodic_.set(i,periodic[i]);',
               'return new DuneType(lowerleft,upperright,elements,periodic_,overlap);'],
              ['"lowerleft"_a', '"upperright"_a', '"elements"_a', '"periodic"_a', '"overlap"_a'])

    gridModule = module(includes, typeName, ctor) # , dynamicAttr=True)

    try:
        lowerleft  = constructor.lower
        upperright = constructor.upper
        elements   = constructor.division
        periodic   = constructor.param.get("periodic", [False,]*dimgrid)
        overlap    = constructor.param.get("overlap", 0)
        return gridModule.HierarchicalGrid(lowerleft,upperright,elements,periodic,overlap).leafView
    except AttributeError:
        return gridModule.reader(constructor).leafView

#################################################################
class TrivialAgglomerate:
    def __init__(self,grid,cubes=None):
        if cubes is None:
            from dune.geometry import quadrilateral, triangle
            try:
                if grid.type == quadrilateral:
                    cubes = True
                elif grid.type == triangle:
                    cubes = False
            except AttributeError:
                cubes = False
                assert False, "Unsupported grid element type for vem space"

        if cubes:
            self.minEdgeNumber = 4
        else:
            self.minEdgeNumber = 3
        self.size = grid.size(0)
    def __call__(self, idx):
        return idx
def trivialAgglomerate(constructor, cubes=False, globalRefine=None, **kwargs):
    if cubes:
        grid = aluCubeGrid(constructor, **kwargs)
    else:
        grid = aluSimplexGrid(constructor, **kwargs)
        if globalRefine is not None:
            grid.hierarchicalGrid.globalRefine(globalRefine)
    agglomerate = TrivialAgglomerate(grid, cubes)
    grid.hierarchicalGrid.agglomerate = agglomerate
    return grid

# http://zderadicka.eu/voronoi-diagrams/
from dune.vem.earcut import EarCut
#from scipy.spatial import Voronoi, voronoi_plot_2d, cKDTree
from scipy.spatial import Delaunay
import numpy

def crossProduct(A):
    x1 = (A[1][0] - A[0][0])
    y1 = (A[1][1] - A[0][1])
    x2 = (A[2][0] - A[0][0])
    y2 = (A[2][1] - A[0][1])
    return (x1*y2 - y1*x2)
def checkConvex(points):
    N = len(points)
    prev = 0
    curr = 0
    for i in range(N):
        temp = [points[i], points[(i + 1) % N], points[(i + 2) % N]]
        curr = crossProduct(temp)
        if (curr != 0):
            if (curr * prev < 0):
                return False
            else:
                prev = curr
    return True

try:
    import triangle
except ImportError:
    triangle = None

from sortedcontainers import SortedDict
import matplotlib.pyplot as plt
class PolyAgglomerate:
    def __init__(self,grid,index):
        indexSet = grid.indexSet
        self.polys = numpy.zeros(grid.size(0),int)
        self.size = 0
        for e in grid.elements:
            idx = index[PolyAgglomerate.roundBary(e.geometry.center)]
            self.polys[indexSet.index(e)] = idx
            self.size = max(self.size,idx+1)

    def __call__(self,idx):
        return self.polys[idx]

    @staticmethod
    def roundBary(a):
        return tuple(round(aa,8) for aa in a)
    @staticmethod
    def construct(cubes,convex,vertices,polygons):
        index = SortedDict()
        if cubes:
            for i,poly in enumerate(polygons):
                assert len(poly)==4
                bary = [sum([vertices[p][i] for p in poly]) / 4 for i in range(2)]
                index[PolyAgglomerate.roundBary(bary)] = i
                polygons[i][2], polygons[i][3] = poly[3], poly[2]
            domain = {"vertices": vertices.astype(float), "cubes": polygons.astype(int)}
        else:
            vertices = numpy.array(vertices)
            tr = []
            earCut = EarCut()
            for nr, p in enumerate(polygons):
                N = len(p)
                if True:
                    tri = earCut(vertices,p)
                elif not convex and not checkConvex(vertices[p]):
                    if not triangle:
                        raise ValueError("""
a grid with non convex polygons requires the 'triangle' package.
Run 'pip install triangle' and try again.
Note that 'convex=False' is the default - if all your polygons are convex
change the parameter to 'True' which will also speedup the grid construction.
""")
                    # use triangle
                    e = [ [p[i],p[(i+1)%N]] for i in range(N) ]
                    domain = { "vertices":vertices, "segments":numpy.array(e) }
                    B = triangle.triangulate(domain,opts="p")
                    tri = B["triangles"]
                    # triangle.compare(plt, domain, B)
                    # plt.show()
                else: # use scipy
                    p = numpy.array(p)
                    poly = numpy.append(p,[p[0]])
                    vert = vertices[p, :]
                    tri = p[Delaunay(vert).simplices]
                tr += [{"triangles":tri}]
                bary = [ (vertices[p0]+vertices[p1]+vertices[p2])/3.
                         for p0,p1,p2 in tr[-1]["triangles"] ]
                for b in bary:
                    index[PolyAgglomerate.roundBary(b)] = nr
            domain = {"vertices":  numpy.array(vertices).astype(numpy.float64),
                      "simplices": numpy.vstack([ t["triangles"] for t in tr ]).astype(int)}
        return domain, index
def polyAgglomerate(constructor,cubes=False,convex=None):
    if convex is None:
        convex = constructor.get("convex",False)
    suffix = "poly"+str(len(constructor["polygons"]))
    domain, index = PolyAgglomerate.construct(cubes,convex,
                                                constructor["vertices"],
                                                constructor["polygons"])
    if cubes:
        grid = aluCubeGrid(domain)
    else:
        grid = aluSimplexGrid(domain)
    agglomerate = PolyAgglomerate(grid,index)
    agglomerate.minEdgeNumber = min([len(p) for p in constructor["polygons"]])
    grid.hierarchicalGrid.agglomerate = agglomerate
    return grid

def polyGrid(constructor,cubes=False, convex=None,**kwargs):
    from dune.alugrid import ALUGridEnvVar
    verbosity = ALUGridEnvVar('ALUGRID_VERBOSITY_LEVEL', 0)
    if isinstance(constructor,dict) and constructor.get("polygons",None) is not None:
        grid = polyAgglomerate(constructor,cubes,convex)
    else:
        grid = trivialAgglomerate(constructor, cubes, **kwargs)
    # in case of a carteisan domain store if old or new boundary ids was used
    # this can be removed in later version - it is only used in dune-fem
    # to give a warning that the boundary ids for the cartesian domains have changed
    try:
        grid.hierarchicalGrid._cartesianConstructionWithIds = constructor.boundaryWasSet
    except AttributeError:
        pass
    return grid

import weakref
agglomerationStorage_ = weakref.WeakKeyDictionary()
def removeAgglo_(agglo):
    for key, value in agglomerationStorage_.items():
        if value == agglo:
            del agglomerationStorage_[key]
            break
def retrieveAgglo_(gv):
    return agglomerationStorage_.get(gv,None)
def insertAgglo_(gv, agglo):
    agglomerationStorage_[gv] = agglo


def plotPolygons(domain):
    figure = plt.figure()
    axe = figure.add_subplot(111)
    for cell in domain["polygons"]:
        polygon = plt.Polygon(
            [ domain["vertices"][i] for i in cell ],
            edgecolor="black",
            facecolor="white"
        )
        axe.add_patch(polygon)
    axe.set_aspect('equal')
    axe.autoscale()
    return figure

try:
    import meshio
    def writePolygons(fileBase, domain, fileFormat="vtu"):
        points = list( domain["vertices"] )
        polys = list( domain["polygons"] )

        # in vtk points are 3d
        points = np.insert(points, 2, 0, axis=1)

        # sort polygons into buckets accoring to equal sizes
        polys.sort(key=len)
        cells = [[polys[0]]]
        for p in polys[1:]:
            if len(p) > len(cells[-1][0]):
                cells += [[p]]
            else:
                cells[-1] += [p]
        cells = [("polygon", c) for c in cells]
        meshio.write_points_cells(fileBase+".vtu", points,
                                  cells, file_format=fileFormat)
except ImportError:
    def writePolygons(fileBase, domain, fileFormat="vtu"):
        raise ImportError("dune.vem.writePolygons requires the meshio package")



# %% [markdown]
# # 2D navier stokes around a cylinder
#
# https://hplgit.github.io/fenics-tutorial/pub/sphinx1/._ftut1004.html#flow-past-a-cylinder
# OMP_NUM_THREADS=4 python cylinderSplit.py

# %%
import matplotlib

matplotlib.rc("image", cmap="jet")
from scipy.sparse import linalg
from math import sqrt
import matplotlib.pyplot as plt
import pygmsh
from dune.ufl import Constant, DirichletBC
import dune.fem.space
import dune.fem.scheme
import dune.vem
from uzawa import dgLaplace, plot

useVem     = True
pressureFV = True
order      = 2

T = 5.0                              # final time
# tau = Constant(T / 10000, "tau")     # fine
tau = Constant(T / 4000, "tau")      # coarse (works with p=2 fails with p=4)
# tau = Constant(T / 5000, "tau")      # coarse (works with p=4)

mu  = Constant(0.001, "mu")          # kinematic viscosity
rho = Constant(1, "rho")             # density

with pygmsh.occ.Geometry() as geom:
    # geom.characteristic_length_max = 0.02
    geom.set_mesh_size_callback( lambda dim, tag, x, y, z, lc:
          min(0.04+(x-0.2)**2+(y-0.2)**2 - 0.1**2, 0.1) )      # coarse
          # min(0.01+(x-0.2)**2+(y-0.2)**2 - 0.1**2, 0.1) )      # fine
    rectangle = geom.add_rectangle([0, 0, 0], 2.2, 0.41)
    cylinder = geom.add_disk([0.2, 0.2, 0.0], 0.05)
    geom.boolean_difference(rectangle, cylinder)

    mesh = geom.generate_mesh()
    points, cells = mesh.points, mesh.cells_dict
    domain = {
        "vertices": points[:, :2].astype(float),
        "simplices": cells["triangle"].astype(int),
    }
    print(len(domain["simplices"]))

gridView = dune.vem.polyGrid(domain)
# gridView.plot()
# plt.show()

if useVem:
    path="splitVemNC_p" + str(order)
    space_u = dune.vem.divFreeSpace( gridView, order=order, conforming=False)
    if pressureFV:
        # space_p = dune.fem.space.finiteVolume( gridView )
        # FIX ERROR:
        space_p = dune.vem.bbdgSpace( gridView, order=0 )
    else:
        space_p = dune.fem.space.lagrange( gridView, order=1 )
    scheme = lambda model, gStab, mStab: dune.vem.vemScheme(
                  model,
                  solver=("suitesparse","umfpack"),
                  # parameters={"newton.verbose":True, "newton.linear.verbose":True},
                  gradStabilization=gStab,
                  massStabilization=mStab )
else:
    path="splitTH_p" + str(order)
    space_u = dune.fem.space.lagrange( gridView, order=order, dimRange=2)
    space_p = dune.fem.space.lagrange( gridView, order=order-1)
    pressureFV = False
    scheme = lambda model, gStab, mStab: dune.fem.scheme.galerkin(
                  model,
                  solver=("suitesparse","umfpack"),
                  # parameters={"newton.verbose":True, "newton.linear.verbose":True}
                  )


# %%
from ufl import (
    FacetNormal,
    Identity,
    SpatialCoordinate,
    TestFunction,
    TrialFunction,
    conditional,
    div,
    dot,
    ds,
    dx,
    inner,
    nabla_grad,
    sqrt,
    sym,
    triangle,
)

u = TrialFunction(space_u)
v = TestFunction(space_u)
p = TrialFunction(space_p)
q = TestFunction(space_p)

u_h = space_u.interpolate([0, 0], name="u_h")
u_h_n = u_h.copy(name="u_h_n")

p_h = space_p.interpolate(0, name="p_h")
p_h_n = p_h.copy(name="p_h_n")

# %%
x = SpatialCoordinate(triangle)

inflow = [6 * x[1] * (0.41 - x[1]) / 0.41**2, 0] # 6->18
dbc_u_in = DirichletBC(space_u, inflow, x[0] <= 1e-8)
dbc_u_noflow = DirichletBC(space_u, [0, 0], None)
dbc_u_out = DirichletBC(space_u, [None,None], 2.2 - x[0] <= 1e-8)

dbc_p_out = DirichletBC(space_p, 0, 2.2 - x[0] <= 1e-8)

# %%
# Define strain-rate tensor
def epsilon(u):
    return sym(nabla_grad(u))

# Define stress tensor
def sigma(u, p):
    return 2 * mu * epsilon(u) - p * Identity(len(u))

# Define expressions used in variational forms
U = 0.5 * (u_h_n + u)
n = FacetNormal(space_u)
f = Constant((0, 0), "f")

# %%

E1 = (
    rho * dot((u - u_h_n) / tau, v) * dx
    + rho * dot(dot(u_h_n, nabla_grad(u_h_n)), v) * dx
    + inner(sigma(U, p_h_n), epsilon(v)) * dx
    # - dot(( mu * nabla_grad(U)-p_h_n ) * n, v) * ds
    - rho * dot(f, v) * dx
) == 0

scheme1 = scheme( [E1, dbc_u_in,dbc_u_out,dbc_u_noflow],
                  gStab=2*mu, mStab=rho/tau )

if pressureFV:
    # dbc_p_out = DirichletBC(space_p, 0, 2.2 - x[0] <= 1e-8)
    laplace_p   = dgLaplace(10, p-p_h_n,q, space_p, 0, conditional(2.2-x[0]<=1e-8,1,0))
    E2 = laplace_p + (1 / tau) * div(u_h_n) * q * dx == 0
    scheme2 = dune.fem.scheme.galerkin(E2,
                # solver="cg",
                solver=("suitesparse","umfpack"),
                # parameters={"newton.linear.verbose":True}
                )
    E3 = (
        dot(u, v) * dx
        - dot(u_h_n, v) * dx
        - tau * (p_h - p_h_n) * div(v) * dx
    ) == 0
    # add bnd term after moving grad(p-p_n)v to (p-p_n)div(v)
    scheme3 = scheme( [E3, dbc_u_in,dbc_u_out,dbc_u_noflow],
                      gStab=None, mStab=1)
else:
    E2 = (
        dot(nabla_grad(p-p_h_n), nabla_grad(q)) * dx
        + (1 / tau) * div(u_h_n) * q * dx
    ) == 0
    scheme2 = dune.fem.scheme.galerkin([E2, dbc_p_out],
                # solver="cg",
                solver=("suitesparse","umfpack"),
                # parameters={"newton.linear.verbose":False}
                )
    E3 = (
        dot(u, v) * dx
        - dot(u_h_n, v) * dx
        + tau * dot(nabla_grad(p_h - p_h_n), v) * dx # can we move to p div(v)
    ) == 0
    scheme3 = scheme( [E3, dbc_u_in,dbc_u_out,dbc_u_noflow],
                      gStab=None, mStab=1)

solve = lambda A,b: linalg.spsolve(A,b)
A1    = dune.fem.operator.linear(scheme1).as_numpy
A1inv = lambda rhs: solve(A1,rhs.as_numpy[:])
A2    = dune.fem.operator.linear(scheme2).as_numpy
A2inv = lambda rhs: solve(A2,rhs.as_numpy[:])
A3    = dune.fem.operator.linear(scheme3).as_numpy
A3inv = lambda rhs: solve(A3,rhs.as_numpy[:])
zero_u = space_u.interpolate([0,0],name="zero")
zero_p = space_p.interpolate(0,name="zero")
rhs_u  = zero_u.copy(name="rhs")
rhs_p  = zero_p.copy(name="rhs")

# %%
from pathlib import Path
p = Path(path)
p.mkdir(exist_ok=True)
vtk = gridView.sequencedVTK(path+f"/2DCylinder",
              subsampling=3,
              pointdata=[p_h], pointvector=[u_h])
vtk()

time = 0
saveStep = 0.05
nextSave = saveStep

while time < T:
    u_h_n.assign(u_h)
    p_h_n.assign(p_h)
    scheme1(zero_u,rhs_u)
    rhs_u.as_numpy[:] *= -1
    scheme1.setConstraints(rhs_u)
    u_h.as_numpy[:] = A1inv(rhs_u)
    u_h_n.assign(u_h)

    scheme2(zero_p,rhs_p)
    rhs_p.as_numpy[:] *= -1
    try:
        scheme2.setConstraints(rhs_p)
    except AttributeError:
        pass
    p_h.as_numpy[:] = A2inv(rhs_p)

    scheme3(zero_u,rhs_u)
    rhs_u.as_numpy[:] *= -1
    scheme3.setConstraints(rhs_u)
    u_h.as_numpy[:] = A3inv(rhs_u)

    time += tau.value
    if time > nextSave:
        print(time, flush=True)
        vtk()
        nextSave += saveStep

# %%



import math, numpy, scipy
from scipy.sparse.linalg import spsolve
from scipy.sparse import coo_matrix, bmat
import dune.fem

# (  M   G  ) ( u     )   ( g ) # g: Dirichlet BC - Neuman BC are natural
# ( -D   0  ) ( sigma ) = ( f )
# sigma = -M^{-1}Gu + M^{-1}g
# f = -Dsigma = DM^{-1}Gu - DM^{-1}g
# DM^{-1}Gu = f + DM^{-1}g
# Au = b with A = DM^{-1}G and b = f + DM^{-1}g
class MixedSolver(scipy.sparse.linalg.LinearOperator):
    def __init__(self, schemeM, schemeG, schemeD):
        self.schemeD = schemeD # needed for rhs
        self.schemeM = schemeM
        self.M = schemeM.linear().as_numpy
        self.invM = scipy.sparse.linalg.splu(self.M.tocsc())
        self.G = schemeG.linear().as_numpy
        self.D = schemeD.linear().as_numpy
        self.tmp = schemeD.domainSpace.interpolate([0,0],name="rhs")
        self.rhs = schemeD.rangeSpace.interpolate(0,name="rhs")
        self.s1 = self.tmp.copy()
        # schemeD.domainSpace.interpolate([0,0],name="tmp").as_numpy[:]
        self.s0 = self.tmp.copy()
        # schemeD.domainSpace.interpolate([0,0],name="tmp").as_numpy[:]
        self.y  = self.rhs.as_numpy.copy()
        # schemeD.rangeSpace.interpolate(0,name="tmp").as_numpy[:]
        self.shape = (schemeD.rangeSpace.size, schemeD.rangeSpace.size)
        self.dtype = self.y.dtype
        self.count = 0

    def update(self, x_coeff, f):
        pass
    def _matvec(self, x_coeff):
        self.s0.as_numpy[:] = self.G@x_coeff[:]
        # self.s1[:] = scipy.sparse.linalg.cg(self.M,self.s0[:])[0]
        # spsolve(self.M,self.s0[:])
        # self.schemeM.solve(rhs=self.s0,target=self.s1)
        self.s1.as_numpy[:] = self.invM.solve(rhs=self.s0.as_numpy[:])
        self.y[:] = self.D@self.s1.as_numpy[:]
        return self.y
    def callback(self,xk):
        self.count += 1
        self.y[:] = self._matvec(xk)
        self.y[:] -= self.rhs.as_numpy[:]
        print(self.count, numpy.dot(self.y,self.y))

    def solve(self,target):
        self.tmp.clear()
        self.schemeD(self.tmp,self.rhs)
        target[0].as_numpy[:] = scipy.sparse.linalg.cg(
          self,self.rhs.as_numpy, atol=0
          # ,callback=lambda xk:self.callback(xk)
          )[0]
        self.tmp.as_numpy[:]  = -self.G@target[0].as_numpy[:]
        self.schemeM.solve(rhs=self.tmp,target=target[1])
        # target[1].as_numpy[:] = spsolve(self.M,self.tmp.as_numpy[:])



import math
import argparse
import random

import dune.fem
import dune.fem.plotting
import dune.grid
import dune.ufl
import dune.vem
import matplotlib.pyplot as plt
import numpy as np
import ufl

dune.fem.threading.use = dune.fem.threading.max

# ************************ #
# Parameters of the script #
# ************************ #
parser = argparse.ArgumentParser(description='2nd order pde using C^1 spaces')
parser.add_argument('-m', type=str, required=True,
                          choices=["C1c","C1nc","C1C0"],
                          help="method (C1c,C1nc,C1C0)")
parser.add_argument('-M', type=str, required=True,
                          choices=["cube","simplex","voronoi","randomcube"],
                          help="mesh (cube,simplex,voronoi,randomcube)")
parser.add_argument('-p', type=str, required=True,
                          choices=["interpolation","poisson","bilaplace","nonvariational","forthorder"],
                          help="pde (interpolation,poisson,bilaplace,nonvariational,forthorder)")
parser.add_argument('-u', type=str, required=True,
                          choices=["u1","u2","u3"],
                          help="solution (u1,u2,u3)")
parser.add_argument('-A', type=str, required=True,
                          default="id",
                          choices=["id","coercive","notH2"],
                          help="solution (id,coercive,notH2)")
parser.add_argument('-l', type=int, required=True, help="polynomial degree (>=2)")
args = parser.parse_args()

method = args.m
gridtype = args.M
pde = args.p
solution_name = args.u
problemA = args.A
order = args.l

useCrissCross = False

# **************** #
# Helper functions #
# **************** #


def H(w):
    return ufl.grad(ufl.grad(w))


def laplace(w):
    return ufl.div(ufl.grad(w))


def bilaplace(w):
    return laplace(laplace(w))


# **************** #
# Main Python code #
# **************** #


def errors(n):
    if gridtype == "voronoi":
        domain = dune.vem.voronoiCells([[0,0],[1,1]], 9*n*n, lloyd=100, load="voronoi")
        cubes = False
    elif gridtype == "randomcube":
        N = 12*n
        h = 1/N
        domain = {"vertices":[], "cubes":[]}
        for i in range(N+1):
            for j in range(N+1):
                if i>0 and i<N:
                    ri = (random.random()-0.5)/10
                else:
                    ri = 0
                if j>0 and j<N:
                    rj = (random.random()-0.5)/10
                else:
                    rj = 0
                # no random perturbation
                # ri,rj = 0,0

                x,y = (j+rj)*h, (i+ri)*h

                # rotate:
                # c,s = np.cos(30*np.pi/180), np.sin(30*np.pi/180)
                # x,y = c*x-s*y, s*x+c*y

                domain["vertices"] += [ [x,y] ]
        for i in range(N):
            for j in range(N):
                domain["cubes"] += [ [ i   *(N+1)+j, i   *(N+1)+j+1,
                                      (i+1)*(N+1)+j,(i+1)*(N+1)+j+1] ]
        cubes = True
    else:
        domain = dune.grid.cartesianDomain([0, 0], [1, 1], [12*n, 12*n])
        cubes = gridtype=="cube"
    grid = dune.vem.polyGrid(domain, cubes=cubes )
    if gridtype == "simplex" and useCrissCross:
        grid.hierarchicalGrid.globalRefine(1)
    # grid.plot()
    """
    indexSet = grid.indexSet
    @dune.grid.gridFunction(grid, name="cells", order=1)
    def polygons(en,x):
        return grid.hierarchicalGrid.agglomerate(indexSet.index(en))
    polygons.plot()
    """

    if method == "C1c":
        testSpaces = [[0, 0], [order - 4, order - 3], [order - 4]]
    elif method == "C1nc":
        testSpaces = [[0], [order - 3, order - 2], [order - 4]]
    elif method == "C1C0":
        testSpaces = [[0], [order - 2, order - 2], [order - 4]]
        # testSpaces = [[0, 0], [order - 4, order - 3], [order - 2]]
        # testSpaces = [0, order - 2, order - 2]
        # testSpaces = [[0], [order - 3, order - 2], [order - 4]]
        # testSpaces = [-1, order - 1, order - 2]

    space = dune.vem.vemSpace(grid, order=order, testSpaces=testSpaces)
    # space = dune.fem.space.lagrange(grid, order=order)

    x = ufl.SpatialCoordinate(space)

    if solution_name == "u1":
        exact = ufl.sin(ufl.pi * x[0])**2 * ufl.sin(ufl.pi * x[1])**2
    elif solution_name == "u2":
        # exact = ufl.sin(ufl.pi * x[0]) ** 2 + ufl.sin(ufl.pi * x[1]) ** 2
        # exact = x[0]*x[1]**2 + (x[0]-0.3)**3
        exact  = ufl.sin(2.7*ufl.pi*x[0])*ufl.cos(2.2*ufl.pi*x[1])
        exact = ( (1-x[0])*x[0]*(1-x[1])*x[1] )**2
    elif solution_name == "u3":
        exact = ufl.sin(2.7*ufl.pi*x[0])*ufl.cos(2.2*ufl.pi*x[1])
    exact = dune.fem.function.uflFunction(grid, name="exact", order=space.order, ufl=exact )

    u      = ufl.TrialFunction(space)
    v      = ufl.TestFunction(space)
    normal = ufl.FacetNormal(space)
    tau    = ufl.as_vector([ -normal[1], normal[0] ])
    hbnd   = ufl.CellVolume(space) / ufl.FacetArea(space)
    hint   = ufl.avg(ufl.CellVolume(space)) / ufl.FacetArea(space)
    beta   = dune.ufl.Constant(10 * space.order**2,"beta")
    dbc    = dune.ufl.DirichletBC(space, exact)

    solution = dune.fem.function.discreteFunction(space, name="solution")
    solution.interpolate(exact)

    if pde == "bilaplace":
        a = ufl.inner(H(u), H(v)) * ufl.dx
        b = bilaplace(exact) * v * ufl.dx
        a += beta/hbnd/hbnd * (u-exact) * v * ufl.ds
        a += beta/hbnd * ufl.dot(ufl.grad(u-exact),normal) *\
                         ufl.dot(ufl.grad(v),normal) * ufl.ds

        scheme = dune.vem.vemScheme(
            [a == b],
            space,
            solver=("suitesparse","umfpack"),
            parameters={"newton.linear.verbose":False,
                        "newton.verbose":False,
                        "newton.linear.preconditioning.method":"jacobi",
                        "newton.linear.tolerance":1e-12,
                        "newton.maxiterations":10},
            hessStabilization=1,
            # boundary="full",
        )
        scheme.solve(target=solution)
    elif pde == "poisson":
        a  = ufl.inner(ufl.grad(u), ufl.grad(v)) * ufl.dx
        b  = -laplace(exact) * v * ufl.dx
        a -= ufl.dot(ufl.grad(u),normal) * v * ufl.ds +\
             ufl.dot(ufl.grad(v),normal) * (u-exact) * ufl.ds
        a += beta/hbnd * (u-exact) * v * ufl.ds
        # a += beta * ufl.dot(ufl.grad(u-exact),tau) * ufl.dot(ufl.grad(v),tau) * ufl.ds

        scheme = dune.vem.vemScheme(
            [a == b],
            solver=("suitesparse","umfpack"),
            gradStabilization=1,
            massStabilization=0,
            parameters={"newton.linear.verbose":False,
                        "newton.verbose":False,
                        "newton.maxiterations":10},
            # boundary=None,
            # boundary="value",
        )
        """
        scheme = dune.fem.scheme.galerkin( a == b, solver=("suitesparse","umfpack") )
        """
        scheme.solve(target=solution)
    elif pde == "nonvariational" or pde == "forthorder":
        if problemA == "id":
            d11=d22 = 1
            d12 = 0
        if problemA == "coercive":
            d11 = 1
            d22 = 1-ufl.ln((x[0]-ufl.sqrt(3)/3)**2+10**(-4))
            d12 = 0
        if problemA == "notH2":
             d11 = 1
             d22 = 2
             d12 = (x[0]*x[1])**(2./3.)
        A = ufl.as_matrix([ [d11,d12], [d12,d22] ])
        # lam = (1+d22)/2 + ufl.sqrt( (1+d22)**2/4 - d22+d12**2 )
        lam = ( d11**2 + d22**2 + 2*d12**2 ) / (d11+d22)
        A /= lam

        if pde == "nonvariational":
            a  = -ufl.inner(A,H(u-exact)) * v * ufl.dx
            a += beta/hbnd * (u-exact) * v * ufl.ds
            # a += beta*lam * ufl.dot(ufl.grad(u-exact),tau) * v * ufl.ds
            # possibly of interest for the non-conforming spaces but does not work at the moment
            # a += beta/hint * ufl.jump(u) * ufl.jump(v) * ufl.dS
            # a += beta/hint * inner( ufl.jump(grad(u)), ufl.jump(grad(v)) ) * ufl.dS
            hS,gS = 0,20   # gs needs to be greater than one for l=4 on simplex grid
            beta.value = 1 # beta=1 seems to be fine
        else:
            a  = ufl.inner(A,H(u-exact)) * laplace(v) * ufl.dx
            a += beta/hbnd**3 * (u-exact) * v * ufl.ds
            hS,gS = 1,0
            beta.value = 1 # beta=1 seems to be fine

        scheme = dune.vem.vemScheme(
            [a == 0],
            solver=("suitesparse","umfpack"),
            hessStabilization=hS,
            gradStabilization=gS,
            massStabilization=0,
            parameters={"newton.linear.verbose":False,
                        "newton.verbose":False,
                        "newton.maxiterations":10},
        )
        """
        scheme = dune.fem.scheme.galerkin( a == b, solver=("suitesparse","umfpack") )
        """
        scheme.solve(target=solution)

    # solution.as_numpy[:] *= interMax / max(solution.as_numpy)
    # for d in solution.as_numpy: print(d,end=" ")
    # print("solution:",max(solution.as_numpy))
    grid.writeVTK(f"{gridtype}{method}{solution_name}{pde}{problemA}{order}_{n}",
                                  celldata={"exact":exact,
                                            "solution":solution,
                                            "dxExact":ufl.grad(exact)[0],
                                            "dxSol":ufl.grad(solution)[0]},
                  subsampling=2)
    # solution.plot()

    error = solution - exact

    error_l2 = math.sqrt(
        dune.fem.function.integrate(grid, ufl.inner(error, error), order=5)
    )
    error_h1 = math.sqrt(
        dune.fem.function.integrate(
            grid, ufl.inner(ufl.grad(error), ufl.grad(error)), order=5
        ))
    error_h2 = math.sqrt(
        dune.fem.function.integrate(grid, ufl.inner(H(error), H(error)), order=5)
    )
    print(n,space.size,"   ",error_l2, error_h1, error_h2)

    return error_l2, error_h1, error_h2, space.size < 50000


def main():
    # N = np.stack([2 * 2 ** np.arange(4), 3 * 2 ** np.arange(4)], axis=-1).flatten()
    N = 2 ** np.arange(7)
    errors_l2 = np.zeros(N.shape)
    errors_h1 = np.zeros(N.shape)
    errors_h2 = np.zeros(N.shape)

    for i, n in enumerate(N):
        errors_l2[i], errors_h1[i], errors_h2[i], ctn = errors(n)
        if not ctn: break
    N = N[:i+1]
    errors_l2 = errors_l2[:i+1]
    errors_h1 = errors_h1[:i+1]
    errors_h2 = errors_h2[:i+1]
    print("# L^2-EOCs:", np.log( errors_l2[1:] / errors_l2[:-1] ) / np.log(0.5) )
    print("# H^1-EOCs:", np.log( errors_h1[1:] / errors_h1[:-1] ) / np.log(0.5) )
    print("# H^2-EOCs:", np.log( errors_h2[1:] / errors_h2[:-1] ) / np.log(0.5) )

    start = lambda p: np.sqrt(errors_h1[0]*errors_l2[0])*N[0]**p
    plt.loglog(N, start(1) / N, "k:", label="order = 1")
    plt.loglog(N, start(2) / N**2, "k--", label="order = 2")
    plt.loglog(N, start(3) / N**3, "k-.", label="order = 3")
    plt.loglog(N, start(4) / N**4, "k-", label="order = 4")

    plt.loglog(N, errors_l2, ".-", label="$L^2$ error")
    plt.loglog(N, errors_h1, ".-", label="$H^1$ error")
    plt.loglog(N, errors_h2, ".-", label="$H^2$ error")

    plt.xlabel("N")
    plt.legend()
    plt.savefig(f"{gridtype}{method}{solution_name}{pde}{problemA}{order}.png",dpi=300)

main()



import dune.vem
import matplotlib
matplotlib.rc( 'image', cmap='jet' )
from matplotlib import pyplot
import math, numpy
from dune.grid import cartesianDomain, gridFunction
from dune.fem.plotting import plotPointData as plot
from dune.fem import integrate
from dune.fem.function import discreteFunction
import dune.fem
from concave import ncGrid

from ufl import *
import dune.ufl

dune.fem.parameter.append({"fem.verboserank": 0})
order = 3
testSpaces = [0,order-2,order-2]

x = SpatialCoordinate(triangle)
massCoeff = 1+sin(dot(x,x))       # factor for mass term
diffCoeff = 1-0.9*cos(dot(x,x))   # factor for diffusion term

def model(space):
    dimR = space.dimRange
    exact = as_vector(dimR*[x[0]*x[1] * cos(pi*x[0]*x[1])])

    u = TrialFunction(space)
    v = TestFunction(space)
    a = (diffCoeff*inner(grad(u),grad(v)) + massCoeff*dot(u,v) ) * dx
    b = sum( [ ( -div(diffCoeff*grad(exact[i])) + massCoeff*exact[i] ) * v[i]
             for i in range(dimR) ] ) * dx
    dbc = [dune.ufl.DirichletBC(space, exact, i+1) for i in range(4)]
    return a,b,dbc,exact

def calc(polyGrid):
    dimR = 1
    orders = (order,order,order-1)
    space = dune.vem.vemSpace( polyGrid, order=orders, storage="numpy",
                               dimRange=dimR,testSpaces=testSpaces)
    a,b,dbc,exact = model(space)
    if False:
        df = space.interpolate(exact,name="solution")
    else:
        df = space.interpolate(dimR*[0],name="solution")
        scheme = dune.vem.vemScheme(
                  [a==b, *dbc], space, solver="cg",
                  parameters={"newton.linear.preconditioner":"sor"},
                  gradStabilization=None, # diffCoeff,
                  massStabilization=massCoeff)
        info = scheme.solve(target=df)
    # df.plot()
    edf = exact-df
    err = [inner(edf,edf),
           inner(grad(edf),grad(edf))]
    return [ numpy.sqrt(e) for e in integrate(err) ]

oldErrors = []
for i in range(1,3):
    errors = []
    N = 5*2**i
    """
    polyGrid = dune.vem.polyGrid(
          dune.vem.voronoiCells([[-0.5,-0.5],[1,1]], N*N, lloyd=100, fileName="test", load=True)
      )
    """
    # polyGrid = dune.vem.polyGrid(cartesianDomain([-0.5,-0.5],[1,1],[N,N]), cubes=False )
    cells = ncGrid(N)
    polyGrid = dune.vem.polyGrid( cells )
    try:
        dune.vem.writePolygons(f"concave_{i}",cells)
    except ImportError: # no meshio
        pass

    errors += calc(polyGrid)

    print(i,polyGrid.hierarchicalGrid.agglomerate.size,"errors:",errors)
    if len(oldErrors)>0:
        eocs = [ math.log(oe/e)/math.log(2.) for oe,e in zip(oldErrors,errors) ]
        print(i,"eocs:", eocs)
        assert eocs[0] > order+0.5
        assert eocs[1] > order-0.5
    oldErrors = errors
